from __future__ import annotations

import os
import socket
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Self

from pydantic import BaseModel

from fmu.settings._resources.log_manager import LogManager
from fmu.settings.models._enums import ChangeType, FilterType
from fmu.settings.models.change_info import ChangeInfo
from fmu.settings.models.log import Filter, Log, LogFileName

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        FMUDirectoryBase,
    )


class ChangelogManager(LogManager[ChangeInfo]):
    """Manages the .fmu changelog file."""

    def __init__(self: Self, fmu_dir: FMUDirectoryBase) -> None:
        """Initializes the Change log resource manager."""
        super().__init__(fmu_dir, Log[ChangeInfo])

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the log file."""
        return Path("logs") / LogFileName.changelog

    def log_update_to_changelog(
        self: Self,
        updates: dict[str, Any],
        old_resource_dict: dict[str, Any],
        relative_path: Path,
    ) -> None:
        """Logs the update of a resource to the changelog."""
        _MISSING_KEY = object()
        for key, new_value in updates.items():
            change_type = ChangeType.update
            if "." in key:
                old_value = self._get_dot_notation_key(
                    resource_dict=old_resource_dict, key=key, default=_MISSING_KEY
                )
            else:
                old_value = old_resource_dict.get(key, _MISSING_KEY)

            if old_value != _MISSING_KEY:
                old_value_string = (
                    str(old_value.model_dump())
                    if isinstance(old_value, BaseModel)
                    else str(old_value)
                )
                new_value_string = (
                    str(new_value.model_dump())
                    if isinstance(new_value, BaseModel)
                    else str(new_value)
                )
                change_string = (
                    f"Updated field '{key}'. Old value: {old_value_string}"
                    f" -> New value: {new_value_string}"
                )
            else:
                change_type = ChangeType.add
                new_value_string = (
                    str(new_value.model_dump())
                    if isinstance(new_value, BaseModel)
                    else str(new_value)
                )
                change_string = f"Added field '{key}'. New value: {new_value_string}"

            change_entry = ChangeInfo(
                timestamp=datetime.now(UTC),
                change_type=change_type,
                user=os.getenv("USER", "unknown"),
                path=self.fmu_dir.path,
                change=change_string,
                hostname=socket.gethostname(),
                file=str(relative_path),
                key=key,
            )
            self.add_log_entry(change_entry)

    def log_merge_to_changelog(
        self: Self, source_path: Path, incoming_path: Path, merged_resources: list[str]
    ) -> None:
        """Logs a change entry with merge details to the changelog."""
        resources_string = ", ".join([f"'{resource}'" for resource in merged_resources])
        change_string = (
            f"Merged resources {resources_string} from "
            f"'{incoming_path}' into '{source_path}'."
        )
        self.add_log_entry(
            ChangeInfo(
                timestamp=datetime.now(UTC),
                change_type=ChangeType.merge,
                user=os.getenv("USER", "unknown"),
                path=source_path,
                change=change_string,
                hostname=socket.gethostname(),
                file=resources_string,
                key=".fmu",
            )
        )

    def _get_latest_change_timestamp(self: Self) -> datetime:
        """Get the timestamp of the latest change entry in the changelog."""
        return self.load()[-1].timestamp

    def get_changelog_diff(
        self: Self, incoming_changelog: ChangelogManager
    ) -> Log[ChangeInfo]:
        """Get new entries from the incoming changelog.

        All log entries from the incoming changelog newer than the
        log entries in the current changelog are returned.
        """
        if self.exists and incoming_changelog.exists:
            starting_point = self._get_latest_change_timestamp()
            return incoming_changelog.filter_log(
                Filter(
                    field_name="timestamp",
                    filter_value=str(starting_point),
                    filter_type=FilterType.date,
                    operator=">",
                )
            )
        raise FileNotFoundError(
            "Changelog resources to diff must exist in both directories: "
            f"Current changelog resource exists: {self.exists}. "
            f"Incoming changelog resource exists: {incoming_changelog.exists}."
        )

    def merge_changelog(
        self: Self, incoming_changelog: ChangelogManager
    ) -> Log[ChangeInfo]:
        """Add new entries from the incoming changelog to the current changelog.

        All log entries from the incoming changelog newer than the
        log entries in the current changelog are added.
        """
        new_log_entries = self.get_changelog_diff(incoming_changelog)
        return self.merge_changes(new_log_entries.root)

    def merge_changes(self: Self, change: list[ChangeInfo]) -> Log[ChangeInfo]:
        """Merge a list of changes into the current changelog.

        All log entries in the change object are added to the changelog.
        """
        for entry in change:
            self.add_log_entry(entry)
        return self.load()



"""Tests specific to the user configuration.

Because the user config and the project config share the same base class a great deal of
the functionality is already tested in test_project_config.py. This test module is more
sparse to handle things specific to the user configuration.
"""

import json
from datetime import UTC, datetime, timedelta

import pytest
from pydantic import SecretStr

from fmu.settings._fmu_dir import UserFMUDirectory
from fmu.settings.models.user_config import UserConfig


def test_set_smda_subscription_key_validates(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests that setting a subscription key is validated by Pydantic."""
    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is None
    with pytest.raises(ValueError, match="Invalid value set"):
        user_fmu_dir.set_config_value("user_api_keys.smda_subscription", 123)
    assert config.user_api_keys.smda_subscription is None

    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", "secret")

    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is not None
    assert config.user_api_keys.smda_subscription.get_secret_value() == "secret"

    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", None)
    config = user_fmu_dir.config.load()


def test_set_smda_subscription_key_writes_to_disk(
    user_fmu_dir: UserFMUDirectory,
) -> None:
    """Tests that setting a subscription key updates to disk."""
    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", "secret")
    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is not None
    assert config.user_api_keys.smda_subscription.get_secret_value() == "secret"
    assert config.user_api_keys.smda_subscription == SecretStr("secret")
    assert user_fmu_dir.get_config_value(
        "user_api_keys.smda_subscription"
    ) == SecretStr("secret")
    assert (
        user_fmu_dir.get_config_value(
            "user_api_keys.smda_subscription"
        ).get_secret_value()
        == "secret"
    )

    with open(user_fmu_dir.config.path, encoding="utf-8") as f:
        config_dict = json.loads(f.read())
    assert config_dict["user_api_keys"]["smda_subscription"] == "secret"


def test_obfuscate_secrets(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests the obfuscate_secrets() method on the UserConfig model."""
    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", "secret")
    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is not None
    assert config.user_api_keys.smda_subscription.get_secret_value() == "secret"
    assert config.user_api_keys.smda_subscription == SecretStr("secret")

    config_dump = config.model_dump()
    assert config_dump["user_api_keys"]["smda_subscription"] == SecretStr("secret")
    config_json = config.model_dump_json()
    assert '{"smda_subscription":"secret"}' in config_json

    secret_config = config.obfuscate_secrets()
    secret_config_dump = secret_config.model_dump()
    assert secret_config_dump["user_api_keys"]["smda_subscription"] == SecretStr(
        "**********"
    )
    secret_config_json = secret_config.model_dump_json()
    assert '{"smda_subscription":"**********"}' in secret_config_json


def test_user_config_reset_has_none_last_modified_at() -> None:
    """Tests that reset() creates a user config with None for last_modified_at."""
    config = UserConfig.reset()

    assert config.last_modified_at is None
    assert config.created_at is not None


def test_user_config_last_modified_at_set_on_save(
    user_fmu_dir: UserFMUDirectory,
) -> None:
    """Tests that last_modified_at is set when saving user config."""
    user_fmu_dir.config.set("cache_max_revisions", 10)

    updated_config = user_fmu_dir.config.load()
    assert updated_config.last_modified_at is not None

    now = datetime.now(UTC)
    one_min_ago = now - timedelta(minutes=1)
    assert one_min_ago <= updated_config.last_modified_at <= now



"""Model for the log entries in the the changelog file."""

from datetime import UTC, datetime
from pathlib import Path

import pandas as pd
from pydantic import AwareDatetime, BaseModel, Field, field_validator

from fmu.settings.models._enums import ChangeType


class ChangeInfo(BaseModel):
    """Represents a change in the changelog file."""

    timestamp: AwareDatetime = Field(
        default_factory=lambda: datetime.now(UTC), strict=True
    )
    change_type: ChangeType
    user: str
    path: Path
    change: str
    hostname: str
    file: str
    key: str

    @field_validator("timestamp", mode="before")
    @classmethod
    def convert_timestamp(cls, value: str) -> AwareDatetime:
        """Convert timestamp values given as a 'str' or Pandas 'Timestamp'.

        Values of other types will be handled by Pydantic.
        """
        if isinstance(value, str):
            return datetime.fromisoformat(value)
        if isinstance(value, pd.Timestamp):
            return value.to_pydatetime()
        return value



"""Tests for mapping models."""

import pytest
from fmu.datamodels.context.mappings import (
    DataSystem,
    RelationType,
    StratigraphyIdentifierMapping,
)
from pydantic import ValidationError

from fmu.settings.models.mappings import MappingGroup


def _make_mapping(
    source_id: str,
    target_id: str,
    relation_type: RelationType,
    source_system: DataSystem = DataSystem.rms,
    target_system: DataSystem = DataSystem.smda,
) -> StratigraphyIdentifierMapping:
    return StratigraphyIdentifierMapping(
        source_system=source_system,
        target_system=target_system,
        relation_type=relation_type,
        source_id=source_id,
        target_id=target_id,
    )


def test_mapping_group_valid_and_display_dict() -> None:
    """MappingGroup groups mappings with a single primary and shared systems."""
    target_id = "Viking GP."
    primary = _make_mapping("Viking Gp", target_id, RelationType.primary)
    alias = _make_mapping("Viking Group", target_id, RelationType.alias)
    equivalent = _make_mapping(target_id, target_id, RelationType.equivalent)

    group = MappingGroup(
        target_id=target_id,
        target_system=DataSystem.smda,
        source_system=DataSystem.rms,
        mappings=[primary, alias, equivalent],
    )

    assert group.primary == primary
    assert group.aliases == [alias]
    assert group.equivalents == [equivalent]
    assert group.to_display_dict() == {
        "official_name": target_id,
        "target_system": "smda",
        "source_system": "rms",
        "primary_source": "Viking Gp",
        "aliases": ["Viking Group"],
        "equivalents": [target_id],
    }


def test_mapping_group_rejects_multiple_primary_mappings() -> None:
    """MappingGroup allows at most one primary mapping."""
    target_id = "Viking GP."
    primary = _make_mapping("Viking Gp", target_id, RelationType.primary)
    primary_two = _make_mapping("Viking Gp 2", target_id, RelationType.primary)
    with pytest.raises(ValidationError, match="at most one primary"):
        MappingGroup(
            target_id=target_id,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, primary_two],
        )


def test_mapping_group_requires_shared_target_id() -> None:
    """All mappings must share the group target_id."""
    primary = _make_mapping("Viking Gp", "Viking GP.", RelationType.primary)
    different_target = _make_mapping("westeros", "Westeros", RelationType.alias)

    with pytest.raises(ValidationError, match="target_id"):
        MappingGroup(
            target_id="Viking GP.",
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, different_target],
        )


def test_mapping_group_requires_shared_target_system() -> None:
    """All mappings must share the group target_system."""
    primary = _make_mapping("Viking Gp", "Viking GP.", RelationType.primary)
    different_target_system = _make_mapping(
        "Viking Gp 2",
        "Viking GP.",
        RelationType.alias,
        target_system=DataSystem.fmu,
    )

    with pytest.raises(ValidationError, match="target_system"):
        MappingGroup(
            target_id="Viking GP.",
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, different_target_system],
        )


def test_mapping_group_requires_shared_source_system() -> None:
    """All mappings must share the group source_system."""
    primary = _make_mapping("Viking Gp", "Viking GP.", RelationType.primary)
    different_source_system = _make_mapping(
        "Viking Gp 2",
        "Viking GP.",
        RelationType.alias,
        source_system=DataSystem.fmu,
    )

    with pytest.raises(ValidationError, match="source_system"):
        MappingGroup(
            target_id="Viking GP.",
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, different_source_system],
        )



"""Tests for the ProjectFMUDirectory class."""

import copy
import inspect
import json
import shutil
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from fmu.datamodels.common.masterdata import Masterdata
from fmu.datamodels.context.mappings import (
    DataSystem,
    RelationType,
    StratigraphyIdentifierMapping,
    StratigraphyMappings,
)
from pytest import MonkeyPatch

from fmu.settings import (
    CacheResource,
    __version__,
    find_nearest_fmu_directory,
    get_fmu_directory,
)
from fmu.settings._fmu_dir import (
    FMUDirectoryBase,
    ProjectFMUDirectory,
    UserFMUDirectory,
)
from fmu.settings._readme_texts import PROJECT_README_CONTENT, USER_README_CONTENT
from fmu.settings._resources.changelog_manager import ChangelogManager
from fmu.settings._resources.lock_manager import DEFAULT_LOCK_TIMEOUT, LockManager
from fmu.settings._resources.mappings_manager import MappingsManager
from fmu.settings.models._enums import ChangeType
from fmu.settings.models.change_info import ChangeInfo
from fmu.settings.models.log import Log
from fmu.settings.models.mappings import Mappings


def test_init_existing_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests initializing an ProjectFMUDirectory on an existing .fmu directory."""
    fmu = ProjectFMUDirectory(fmu_dir.base_path)
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path


def test_get_fmu_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests initializing an ProjectFMUDirectory via get_fmu_directory."""
    fmu = get_fmu_directory(fmu_dir.base_path)
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path


def test_find_nearest_fmu_directory(
    monkeypatch: MonkeyPatch, fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests initializing an ProjectFMUDirectory via find_nearest_fmu_directory."""
    subdir = fmu_dir.path / "subdir"
    subdir.mkdir()
    subdir2 = fmu_dir.path / "subdir2"
    subdir2.mkdir()
    subsubdir = subdir / "subsubdir"
    subsubdir.mkdir()

    fmu = find_nearest_fmu_directory(str(subsubdir))
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path

    monkeypatch.chdir(fmu_dir.base_path)
    fmu = find_nearest_fmu_directory()
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path

    monkeypatch.chdir(subdir2)
    fmu = find_nearest_fmu_directory()
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path


def test_init_on_missing_directory(tmp_path: Path) -> None:
    """Tests initializing with a missing directory raises."""
    with pytest.raises(
        FileNotFoundError, match=f"No .fmu directory found at {tmp_path}"
    ):
        ProjectFMUDirectory(tmp_path)


def test_init_when_fmu_is_not_a_directory(tmp_path: Path) -> None:
    """Tests initialized on a .fmu non-directory raises."""
    (tmp_path / ".fmu").touch()
    with pytest.raises(
        FileExistsError, match=f".fmu exists at {tmp_path} but is not a directory"
    ):
        ProjectFMUDirectory(tmp_path)


def test_find_fmu_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests find_fmu_directory method on nested children."""
    child = fmu_dir.base_path / "child"
    grand_child = child / "grandchild"
    grand_child.mkdir(parents=True)

    found_dir = ProjectFMUDirectory.find_fmu_directory(grand_child)
    assert found_dir == fmu_dir.path


def test_find_fmu_directory_not_found(tmp_path: Path) -> None:
    """Tests find_fmu_directory() returns None if no .fmu found."""
    empty_dir = tmp_path / "empty"
    empty_dir.mkdir()

    found_dir = ProjectFMUDirectory.find_fmu_directory(empty_dir)
    assert found_dir is None


def test_find_nearest(fmu_dir: ProjectFMUDirectory) -> None:
    """Test find_nearest factory method."""
    subdir = fmu_dir.base_path / "subdir"
    subdir.mkdir()

    fmu = ProjectFMUDirectory.find_nearest(subdir)
    assert fmu.path == fmu_dir.path


def test_find_nearest_not_found(tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
    """Test find_nearest raises FileNotFoundError when not found."""
    monkeypatch.chdir(tmp_path)
    with pytest.raises(
        FileNotFoundError, match=f"No .fmu directory found at or above {tmp_path}"
    ):
        ProjectFMUDirectory.find_nearest()
    with pytest.raises(
        FileNotFoundError, match=f"No .fmu directory found at or above {tmp_path}"
    ):
        ProjectFMUDirectory.find_nearest(tmp_path)


def test_cache_property_returns_cached_manager(fmu_dir: ProjectFMUDirectory) -> None:
    """Cache manager should be memoized and ready for use."""
    cache = fmu_dir.cache

    assert cache is fmu_dir.cache
    assert fmu_dir._cache_manager is cache
    assert cache.max_revisions == 5  # noqa: PLR2004


def test_set_cache_max_revisions_updates_manager(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Changing retention should update the existing cache manager."""
    cache = fmu_dir.cache
    fmu_dir.cache_max_revisions = 7

    assert cache.max_revisions == 7  # noqa: PLR2004


def test_changelog_property_returns_changelog_manager(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Changelog manager should be memoized and ready for use."""
    assert isinstance(fmu_dir.changelog, ChangelogManager)
    assert fmu_dir.changelog is fmu_dir._changelog


def test_mappings_property_returns_mappings_manager(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Mappings manager should be memoized and ready for use."""
    assert isinstance(fmu_dir.mappings, MappingsManager)
    assert fmu_dir.mappings is fmu_dir._mappings


def test_get_cache_content_returns_model(fmu_dir: ProjectFMUDirectory) -> None:
    """Cache content should be returned as a validated model."""
    cached_config = fmu_dir.config.load().model_copy(update={"version": "1.2.3"})
    snapshot = fmu_dir.cache.store_revision(
        CacheResource.config.value,
        cached_config.model_dump_json(by_alias=True, indent=2),
    )
    assert snapshot is not None

    cache_content = fmu_dir.get_cache_content(
        CacheResource.config.value,
        snapshot.name,
    )

    assert cache_content.model_dump() == cached_config.model_dump()


def test_get_cache_content_raises_for_unsupported_resource(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Unsupported cache resources should raise."""
    with pytest.raises(ValueError, match="not supported for cache operations"):
        fmu_dir.get_cache_content("unsupported.json", "missing.json")


def test_restore_from_cache_restores_mappings(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Restoring mappings should refresh the resource cache."""
    cached_mapping = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.primary,
        source_id="TopVolantis",
        target_id="VOLANTIS GP. Top",
    )
    cached_mappings = Mappings(stratigraphy=StratigraphyMappings(root=[cached_mapping]))
    snapshot = fmu_dir.cache.store_revision(
        CacheResource.mappings.value,
        cached_mappings.model_dump_json(by_alias=True, indent=2),
    )
    assert snapshot is not None

    current_mapping = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.primary,
        source_id="TopTherys",
        target_id="THERYS GP. Top",
    )
    current_mappings = Mappings(
        stratigraphy=StratigraphyMappings(root=[current_mapping])
    )
    fmu_dir.mappings.save(current_mappings)

    fmu_dir.restore_from_cache(CacheResource.mappings.value, snapshot.name)

    assert fmu_dir.mappings._cache is not None
    assert fmu_dir.mappings._cache.model_dump() == cached_mappings.model_dump()
    restored = fmu_dir.mappings.load()
    assert restored.model_dump() == cached_mappings.model_dump()


def test_restore_from_cache_raises_for_unsupported_resource(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Unsupported cache resources should raise on restore."""
    with pytest.raises(ValueError, match="not supported for cache restoration"):
        fmu_dir.restore_from_cache("unsupported.json", "missing.json")


def test_get_config_value(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests get_config_value retrieves correctly from the config."""
    assert fmu_dir.get_config_value("version") == __version__
    assert fmu_dir.get_config_value("created_by") == "user"


def test_set_config_value(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests set_config_value sets and writes the result."""
    fmu_dir.set_config_value("version", "200.0.0")
    with open(fmu_dir.config.path, encoding="utf-8") as f:
        config_dict = json.loads(f.read())

    assert config_dict["version"] == "200.0.0"
    assert fmu_dir.get_config_value("version") == "200.0.0"
    assert fmu_dir.config.load().version == "200.0.0"


def test_update_config(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests update_config updates and saves the config for multiple values."""
    updated_config = fmu_dir.update_config({"version": "2.0.0", "created_by": "user2"})

    assert updated_config.version == "2.0.0"
    assert updated_config.created_by == "user2"

    assert fmu_dir.config.load() is not None
    assert fmu_dir.get_config_value("version", None) == "2.0.0"
    assert fmu_dir.get_config_value("created_by", None) == "user2"

    config_file = fmu_dir.config.path
    with open(config_file, encoding="utf-8") as f:
        saved_config = json.load(f)

    assert saved_config["version"] == "2.0.0"
    assert saved_config["created_by"] == "user2"


def test_update_config_invalid_data(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that update_config raises ValidationError on bad data."""
    updates = {"version": 123}
    with pytest.raises(
        ValueError,
        match=f"Invalid value set for 'ProjectConfigManager' with updates '{updates}'",
    ):
        fmu_dir.update_config(updates)


def test_get_file_path(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests get_file_path returns correct path."""
    path = fmu_dir.get_file_path("test.txt")
    assert path == fmu_dir.path / "test.txt"


def test_file_exists(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests file_exists returns correct boolean."""
    test_file = fmu_dir.path / "exists.txt"
    test_file.touch()

    assert fmu_dir.file_exists("exists.txt") is True
    assert fmu_dir.file_exists("doesnt.txt") is False


def test_read_file(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests read_file reads bytes correctly."""
    test_file = fmu_dir.path / "bin.dat"
    test_data = b"test bin data"
    test_file.write_bytes(test_data)

    data = fmu_dir.read_file("bin.dat")
    assert data == test_data


def test_read_file_not_found(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests read_file raises FileNotFoundError for missing files."""
    with pytest.raises(FileNotFoundError, match="No such file or directory"):
        fmu_dir.read_file("not_real.txt")


def test_read_text_file(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests read_text_file reads text correctly."""
    test_file = fmu_dir.path / "text.txt"
    test_text = "test text data å"
    test_file.write_text(test_text)

    text = fmu_dir.read_text_file("text.txt")
    assert text == test_text


def test_write_text_file(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests write_text_file writes text correctly."""
    test_text = "new text data æ"
    fmu_dir.write_text_file("new_text.txt", test_text)

    file_path = fmu_dir.path / "new_text.txt"
    assert file_path.exists()
    assert file_path.read_text() == test_text


def test_write_file_creates_dir(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests write_file creates parent directories."""
    test_data = b"nested data"
    fmu_dir.write_file("nested/dir/file.dat", test_data)

    nested_dir = fmu_dir.path / "nested" / "dir"
    assert nested_dir.is_dir()

    file_path = nested_dir / "file.dat"
    assert file_path.exists()
    assert file_path.read_bytes() == test_data


def test_write_operations_raise_when_locked(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests write helpers raise when another process holds the lock."""
    lock = LockManager(fmu_dir)
    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.acquire()

    with pytest.raises(PermissionError, match="Cannot write to .fmu directory"):
        fmu_dir.write_text_file("blocked.txt", "blocked")

    with pytest.raises(PermissionError, match="Cannot write to .fmu directory"):
        fmu_dir.write_file("blocked.bin", b"blocked")

    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.release()


def test_list_files(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that list_files returns the correct files."""
    (fmu_dir.path / "file1.txt").touch()
    (fmu_dir.path / "file2.txt").touch()

    subdir = fmu_dir.path / "subdir"
    subdir.mkdir()
    (subdir / "file3.txt").touch()

    files = fmu_dir.list_files()
    filenames = [f.name for f in files]

    assert "file1.txt" in filenames
    assert "file2.txt" in filenames
    assert "config.json" in filenames

    assert "file3.txt" not in filenames

    subdir_files = fmu_dir.list_files("subdir")
    assert len(subdir_files) == 1
    assert subdir_files[0].name == "file3.txt"

    not_subdir_files = fmu_dir.list_files("not_subdir")
    assert not_subdir_files == []


def test_ensure_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that ensure_directory creates directories."""
    dir_path = fmu_dir.ensure_directory("nested/test/dir")
    assert dir_path.exists()
    assert dir_path.is_dir()


def test_user_init_existing_directory(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests initializing an ProjectFMUDirectory on an existing .fmu directory."""
    with patch("pathlib.Path.home", return_value=user_fmu_dir.base_path):
        fmu = UserFMUDirectory()

    assert fmu.path == user_fmu_dir.path
    assert fmu.base_path == user_fmu_dir.base_path


def test_user_init_on_missing_directory(tmp_path: Path) -> None:
    """Tests initializing with a missing directory raises."""
    with (
        patch("pathlib.Path.home", return_value=tmp_path),
        pytest.raises(
            FileNotFoundError, match=f"No .fmu directory found at {tmp_path}"
        ),
    ):
        UserFMUDirectory()


def test_user_init_when_fmu_is_not_a_directory(tmp_path: Path) -> None:
    """Tests initialized on a .fmu non-directory raises."""
    (tmp_path / ".fmu").touch()
    with (
        patch("pathlib.Path.home", return_value=tmp_path),
        pytest.raises(
            FileExistsError, match=f".fmu exists at {tmp_path} but is not a directory"
        ),
    ):
        UserFMUDirectory()


def test_update_user_config(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests update_config updates and saves the user config for multiple values."""
    recent_dir = "/foo/bar"
    updated_config = user_fmu_dir.update_config(
        {"version": "2.0.0", "recent_project_directories": [recent_dir]}
    )

    assert updated_config.version == "2.0.0"
    assert updated_config.recent_project_directories == [Path(recent_dir)]

    assert user_fmu_dir.config.load() is not None
    assert user_fmu_dir.get_config_value("version", None) == "2.0.0"
    assert user_fmu_dir.get_config_value("recent_project_directories") == [
        Path(recent_dir)
    ]

    config_file = user_fmu_dir.config.path
    with open(config_file, encoding="utf-8") as f:
        saved_config = json.load(f)

    assert saved_config["version"] == "2.0.0"
    assert saved_config["recent_project_directories"] == [recent_dir]


def test_update_user_config_invalid_data(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests that update_config raises ValidationError on bad data."""
    updates = {"recent_project_directories": [123]}
    with pytest.raises(
        ValueError,
        match="Invalid value set for 'UserConfigManager' with updates "
        "'{'recent_project_directories':",
    ):
        user_fmu_dir.update_config(updates)


def test_update_user_config_non_unique_recent_projects(
    user_fmu_dir: UserFMUDirectory,
) -> None:
    """Tests that update_config raises on non-unique recent_project_directories."""
    updates = {"recent_project_directories": [Path("/foo/bar"), Path("/foo/bar")]}
    with pytest.raises(ValueError, match="unique entries"):
        user_fmu_dir.update_config(updates)


def test_acquire_lock_on_project_fmu(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that a lock can be acquired on the project dir."""
    fmu_dir._lock.acquire()
    assert fmu_dir._lock.is_acquired()
    assert fmu_dir._lock.exists
    assert (fmu_dir.path / ".lock").exists()


def test_acquire_lock_on_user_fmu(
    user_fmu_dir: UserFMUDirectory,
) -> None:
    """Tests that a lock can be acquired on the user dir."""
    user_fmu_dir._lock.acquire()
    assert user_fmu_dir._lock.is_acquired()
    assert user_fmu_dir._lock.exists
    assert (user_fmu_dir.path / ".lock").exists()


def test_restore_rebuilds_project_fmu_from_cache(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that restore should recreate missing files using cached config data."""
    fmu_dir.update_config({"version": "123.4.5"})
    cached_dump = json.loads((fmu_dir.path / "config.json").read_text())

    shutil.rmtree(fmu_dir.path)
    assert not fmu_dir.path.exists()

    fmu_dir.restore()

    assert fmu_dir.path.exists()
    readme_path = fmu_dir.path / "README"
    assert readme_path.exists()
    assert readme_path.read_text() == PROJECT_README_CONTENT

    restored_dump = json.loads((fmu_dir.path / "config.json").read_text())

    cached_dump.pop("last_modified_at", None)
    restored_dump.pop("last_modified_at", None)
    assert restored_dump == cached_dump

    cache_dir = fmu_dir.path / "cache" / "config"
    assert cache_dir.is_dir()
    assert any(cache_dir.iterdir())


def test_restore_resets_when_cache_missing(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that restore should fall back to reset when no cached config exists."""
    fmu_dir.config._cache = None
    shutil.rmtree(fmu_dir.path)
    assert not fmu_dir.path.exists()

    with patch.object(
        fmu_dir.config, "reset", wraps=fmu_dir.config.reset
    ) as mock_reset:
        fmu_dir.restore()

    mock_reset.assert_called_once()
    assert fmu_dir.path.exists()
    readme_path = fmu_dir.path / "README"
    assert readme_path.exists()
    assert readme_path.read_text() == PROJECT_README_CONTENT
    assert (fmu_dir.config.path).exists()


def test_restore_rebuilds_user_fmu(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests that user FMU restore should recreate missing files using cached state."""
    cached_dump = json.loads((user_fmu_dir.path / "config.json").read_text())

    shutil.rmtree(user_fmu_dir.path)
    assert not user_fmu_dir.path.exists()

    user_fmu_dir.restore()

    assert user_fmu_dir.path.exists()
    readme_path = user_fmu_dir.path / "README"
    assert readme_path.exists()
    assert readme_path.read_text() == USER_README_CONTENT

    restored_dump = json.loads((user_fmu_dir.path / "config.json").read_text())

    cached_dump.pop("last_modified_at", None)
    restored_dump.pop("last_modified_at", None)
    assert restored_dump == cached_dump


def test_fmu_directory_base_exposes_lock_timeout_kwarg() -> None:
    """Tests that the kw-only lock timeout argument remains available."""
    signature = inspect.signature(FMUDirectoryBase.__init__)
    lock_timeout = signature.parameters.get("lock_timeout_seconds")

    assert lock_timeout is not None, "lock_timeout_seconds kwarg missing from base init"
    assert lock_timeout.kind is inspect.Parameter.KEYWORD_ONLY
    assert lock_timeout.default == DEFAULT_LOCK_TIMEOUT


def test_find_rms_projects_none_found(fmu_dir: ProjectFMUDirectory) -> None:
    """Test finding RMS projects when none exist."""
    projects = fmu_dir.find_rms_projects()
    assert projects == []


def test_find_rms_projects_single(fmu_dir: ProjectFMUDirectory) -> None:
    """Test finding a single RMS project."""
    rms_project = fmu_dir.base_path / "rms" / "model" / "test.rms"
    rms_project.mkdir(parents=True)
    (rms_project / ".master").write_text("master content")
    (rms_project / "rms.ini").write_text("[RMS]")

    projects = fmu_dir.find_rms_projects()

    assert len(projects) == 1
    assert projects[0] == rms_project


def test_find_rms_projects_multiple(fmu_dir: ProjectFMUDirectory) -> None:
    """Test finding multiple RMS projects."""
    rms_project_1 = fmu_dir.base_path / "rms" / "model" / "drogon.rms14.2.2"
    rms_project_1.mkdir(parents=True)
    (rms_project_1 / ".master").write_text("master content")
    (rms_project_1 / "rms.ini").write_text("[RMS]")

    rms_project_2 = fmu_dir.base_path / "rms" / "model" / "drogon.rms14.1.0"
    rms_project_2.mkdir(parents=True)
    (rms_project_2 / ".master").write_text("master content")
    (rms_project_2 / "rms.ini").write_text("[RMS]")

    projects = fmu_dir.find_rms_projects()

    assert len(projects) == 2  # noqa: PLR2004
    assert set(projects) == {rms_project_1, rms_project_2}


def test_find_rms_projects_incomplete_missing_master(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Test that projects missing .master file are not included."""
    incomplete = fmu_dir.base_path / "rms" / "model" / "incomplete"
    incomplete.mkdir(parents=True)
    (incomplete / "rms.ini").write_text("[RMS]")

    projects = fmu_dir.find_rms_projects()
    assert projects == []


def test_find_rms_projects_incomplete_missing_ini(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Test that projects missing rms.ini file are not included."""
    incomplete = fmu_dir.base_path / "rms" / "model" / "incomplete"
    incomplete.mkdir(parents=True)
    (incomplete / ".master").write_text("master content")

    projects = fmu_dir.find_rms_projects()
    assert projects == []


def test_find_rms_projects_ignores_files_in_model_dir(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Test that extra files in rms/model do not affect detection."""
    model_dir = fmu_dir.base_path / "rms" / "model"
    model_dir.mkdir(parents=True)
    (model_dir / "not_a_directory.txt").write_text("some file")
    rms_project = model_dir / "valid.rms"
    rms_project.mkdir()
    (rms_project / ".master").write_text("content")
    (rms_project / "rms.ini").write_text("[RMS]")

    projects = fmu_dir.find_rms_projects()

    assert len(projects) == 1
    assert projects[0] == rms_project


def test_find_rms_projects_ignores_rms_model_as_file(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Test that rms/model path that is a file is ignored."""
    rms_dir = fmu_dir.base_path / "rms"
    rms_dir.mkdir()
    (rms_dir / "model").write_text("not a directory")

    project = fmu_dir.find_rms_projects()

    assert len(project) == 0


def test_find_rms_projects_ignores_model_directory_in_subfolders_of_project_root(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Test that rms/model under subfolders of the root is ignored."""
    invalid_model_location = fmu_dir.base_path / "subfolder" / "rms" / "model"
    invalid_model_location.mkdir(parents=True)
    (invalid_model_location / ".master").write_text("content")
    (invalid_model_location / "rms.ini").write_text("[RMS]")

    projects = fmu_dir.find_rms_projects()

    assert len(projects) == 0


def test_find_rms_projects_with_config_path(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Test storing and retrieving RMS project path in config."""
    rms_project = fmu_dir.base_path / "rms" / "model" / "my.rms"
    rms_project.mkdir(parents=True)
    (rms_project / ".master").write_text("content")
    (rms_project / "rms.ini").write_text("[RMS]")

    fmu_dir.set_config_value("rms", {"path": str(rms_project), "version": "14.2.2"})

    stored_rms = fmu_dir.get_config_value("rms")
    assert stored_rms.path == rms_project

    fmu_dir2 = ProjectFMUDirectory(fmu_dir.base_path)
    assert fmu_dir2.get_config_value("rms").path == rms_project


def test_fmu_directory_base_get_dir_diff_with_other_fmu_dir(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    masterdata_dict: dict[str, Any],
) -> None:
    """Tests getting the resource diff with another .fmu directory."""
    new_fmu_dir = extra_fmu_dir
    new_fmu_dir.set_config_value("masterdata", masterdata_dict)
    diff_fmu_dir = fmu_dir.get_dir_diff(new_fmu_dir)

    assert len(diff_fmu_dir) == 1
    assert "config" in diff_fmu_dir
    assert len(diff_fmu_dir["config"]) == 1
    assert diff_fmu_dir["config"][0][0] == "masterdata"
    assert diff_fmu_dir["config"][0][1] is None
    assert diff_fmu_dir["config"][0][2] == Masterdata.model_validate(masterdata_dict)


def test_fmu_directory_base_get_dir_diff_only_diff_whitelisted_resources(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    masterdata_dict: dict[str, Any],
) -> None:
    """Tests that only the whitelisted resources are diffed."""
    new_fmu_dir = extra_fmu_dir
    new_fmu_dir.set_config_value("masterdata", masterdata_dict)
    new_fmu_dir._lock.acquire()

    assert new_fmu_dir._changelog.exists
    assert new_fmu_dir._lock.exists
    assert not fmu_dir._changelog.exists
    assert not fmu_dir._lock.exists

    diff_fmu_dir = fmu_dir.get_dir_diff(new_fmu_dir)

    assert len(diff_fmu_dir) == 1
    assert diff_fmu_dir["config"]
    assert "_lock" not in diff_fmu_dir


def test_fmu_directory_base_get_dir_diff_skip_when_resource_not_exist(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that sync is skipped for resources that does not exist.

    When a resource does not exist in one of the .fmu directories,
    sync is skipped for that resource.
    """
    new_fmu_dir = extra_fmu_dir
    new_fmu_dir.config.path.unlink()

    diff_fmu_dir = fmu_dir.get_dir_diff(new_fmu_dir)
    assert len(diff_fmu_dir) == 0


def test_fmu_directory_base_sync_dir_with_other_fmu_dir(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    masterdata_dict: dict[str, Any],
) -> None:
    """Tests syncing resources with another .fmu directory."""
    new_fmu_dir = extra_fmu_dir
    new_fmu_dir.set_config_value("masterdata", masterdata_dict)

    assert new_fmu_dir.config.load().masterdata
    assert not fmu_dir.config.load().masterdata

    updated_resources = fmu_dir.sync_dir(new_fmu_dir)

    assert len(updated_resources) == 1
    assert "config" in updated_resources
    assert updated_resources["config"].masterdata == Masterdata.model_validate(
        masterdata_dict
    )
    assert new_fmu_dir.config.load().masterdata == fmu_dir.config.load().masterdata


def test_fmu_directory_base_sync_dir_only_whitelisted_resources(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    masterdata_dict: dict[str, Any],
) -> None:
    """Tests that only the whitelisted resources are synced."""
    new_fmu_dir = extra_fmu_dir
    new_fmu_dir.set_config_value("masterdata", masterdata_dict)
    new_fmu_dir._lock.acquire()

    assert new_fmu_dir._changelog.exists
    assert new_fmu_dir._lock.exists
    assert not fmu_dir._changelog.exists
    assert not fmu_dir._lock.exists

    updates = fmu_dir.sync_dir(new_fmu_dir)

    assert "_lock" not in updates
    assert len(updates) == 1
    assert updates["config"]
    assert not fmu_dir._lock.exists


def test_fmu_directory_base_sync_dir_skip_when_resource_not_exist(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that sync is skipped for resources that does not exist.

    When a resource does not exist in one of the .fmu directories,
    sync is skipped for that resource.
    """
    new_fmu_dir = extra_fmu_dir
    new_fmu_dir.config.path.unlink()
    updates = fmu_dir.sync_dir(new_fmu_dir)
    assert len(updates) == 0


def test_fmu_directory_base_sync_dir_skip_when_no_diff(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that sync is skipped for resources with no diff.

    When doing a sync of two equal resources, sync is skipped for that resource.
    """
    new_fmu_dir = extra_fmu_dir
    assert fmu_dir.config.exists
    assert new_fmu_dir.config.exists
    assert fmu_dir.config.load() == new_fmu_dir.config.load()

    updates = fmu_dir.sync_dir(new_fmu_dir)
    assert len(updates) == 0


def test_fmu_directory_base_sync_runtime_variables(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that runtime variables are synced as part of .fmu directory sync.

    When resources in the .fmu directory are synced,
    runtime variables should be synced accordingly.
    """
    new_fmu_dir = extra_fmu_dir
    new_cache_max_revisions = 10
    old_cache_max_revisions = fmu_dir.config.load().cache_max_revisions
    assert fmu_dir.cache_max_revisions != new_cache_max_revisions
    assert fmu_dir.cache_max_revisions == fmu_dir.config.load().cache_max_revisions

    new_config = new_fmu_dir.config.load()
    new_config.cache_max_revisions = new_cache_max_revisions
    new_fmu_dir.config.save(new_config)

    updates = fmu_dir.sync_dir(new_fmu_dir)
    assert "config" in updates
    assert updates["config"].cache_max_revisions == new_cache_max_revisions
    assert fmu_dir.cache_max_revisions == new_cache_max_revisions
    assert fmu_dir.cache.max_revisions == new_cache_max_revisions

    config = fmu_dir.config.load()
    assert config.cache_max_revisions is new_cache_max_revisions
    config.cache_max_revisions = old_cache_max_revisions
    fmu_dir.config.save(config)
    assert (
        fmu_dir.config.load(force=True).cache_max_revisions == old_cache_max_revisions
    )
    assert fmu_dir.cache_max_revisions == new_cache_max_revisions
    assert fmu_dir.cache.max_revisions == new_cache_max_revisions

    fmu_dir._sync_runtime_variables()

    assert fmu_dir.cache_max_revisions == old_cache_max_revisions
    assert fmu_dir.cache.max_revisions == old_cache_max_revisions


def test_fmu_directory_base_get_dir_diff_with_same_changelog(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests getting the changelog diff with equivalent directories."""
    log_entry = ChangeInfo(
        change_type=ChangeType.add,
        user="fmu_dir_user",
        path=Path("/some_path"),
        change="some change",
        hostname="host",
        file="config",
        key="test_key",
    )
    fmu_dir._changelog.add_log_entry(log_entry)
    extra_fmu_dir._changelog.add_log_entry(log_entry)
    dir_diff = fmu_dir.get_dir_diff(extra_fmu_dir)

    for resource, change_list in dir_diff.items():
        assert len(change_list) == 0, f"Resource {resource} has changes"


def test_fmu_directory_base_get_dir_diff_with_changelog(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests getting the changelog diff with another .fmu directory."""
    log_entry = ChangeInfo(
        change_type=ChangeType.add,
        user="fmu_dir_user",
        path=Path("/some_path"),
        change="some change",
        hostname="host",
        file="config",
        key="test_key",
    )
    fmu_dir._changelog.add_log_entry(log_entry)
    assert len(fmu_dir._changelog.load()) == 1
    assert fmu_dir._changelog.load()[0] == log_entry

    new_fmu_dir = extra_fmu_dir
    new_log_entry = ChangeInfo(
        change_type=ChangeType.add,
        user="new_fmu_dir_user",
        path=Path("/some_new_path"),
        change="new change",
        hostname="new_host",
        file="config",
        key="new_test_key",
    )
    new_fmu_dir._changelog.add_log_entry(new_log_entry)
    assert len(new_fmu_dir._changelog.load()) == 1
    assert new_fmu_dir._changelog.load()[0] == new_log_entry

    dir_diff = fmu_dir.get_dir_diff(new_fmu_dir)

    expected_no_of_resources = 2
    assert len(dir_diff) == expected_no_of_resources
    assert dir_diff["config"] == []
    assert "_changelog" in dir_diff
    changelog_diff = dir_diff["_changelog"]

    assert len(changelog_diff) == 1
    assert changelog_diff[0][0] == "changelog"
    assert changelog_diff[0][1] is None
    assert isinstance(changelog_diff[0][2], Log)
    assert len(changelog_diff[0][2]) == 1
    assert changelog_diff[0][2].root[0] == new_log_entry


def test_fmu_directory_base_sync_dir_with_changelog(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests syncing changelog with another .fmu directory."""
    log_entry = ChangeInfo(
        change_type=ChangeType.add,
        user="fmu_dir_user",
        path=Path("/some_path"),
        change="some change",
        hostname="host",
        file="config",
        key="test_key",
    )
    fmu_dir._changelog.add_log_entry(log_entry)
    assert len(fmu_dir._changelog.load()) == 1
    assert fmu_dir._changelog.load()[0] == log_entry

    new_fmu_dir = extra_fmu_dir
    new_log_entry = ChangeInfo(
        change_type=ChangeType.add,
        user="new_fmu_dir_user",
        path=Path("/some_new_path"),
        change="new change",
        hostname="new_host",
        file="config",
        key="new_test_key",
    )
    new_fmu_dir._changelog.add_log_entry(new_log_entry)
    assert len(new_fmu_dir._changelog.load()) == 1
    assert new_fmu_dir._changelog.load()[0] == new_log_entry

    updated_resources = fmu_dir.sync_dir(new_fmu_dir)

    assert len(updated_resources) == 1
    assert "_changelog" in updated_resources
    updated_changelog: Log[ChangeInfo] = updated_resources["_changelog"]
    expected_log_length = 3
    assert len(updated_changelog) == expected_log_length
    assert updated_changelog[0] == log_entry
    assert updated_changelog[1] == new_log_entry
    assert updated_changelog[2].change_type == ChangeType.merge
    assert "_changelog" in updated_changelog[2].change
    assert updated_changelog[2].path == fmu_dir.path


def test_fmu_directory_base_sync_dir_with_all_resources(
    fmu_dir: ProjectFMUDirectory,
    masterdata_dict: dict[str, Any],
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests syncing with another .fmu directory with config, mappings and changelog."""
    assert fmu_dir.config.load().masterdata is None
    fmu_dir.set_config_value("masterdata", masterdata_dict)
    assert fmu_dir.config.load().masterdata is not None
    assert len(fmu_dir._changelog.load()) == 1
    assert fmu_dir._changelog.load()[0].change_type == ChangeType.update

    fmu_dir._mappings.update_stratigraphy_mappings(StratigraphyMappings(root=[]))
    new_fmu_dir = extra_fmu_dir
    new_strat_mapping = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.primary,
        source_id="TopViking",
        target_id="VIKING GP. Top",
    )
    new_fmu_dir._mappings.update_stratigraphy_mappings(
        StratigraphyMappings(root=[new_strat_mapping])
    )

    assert len(new_fmu_dir._changelog.load()) == 1
    assert new_fmu_dir._changelog.load()[0].key == "stratigraphy"
    assert new_fmu_dir._changelog.load()[0].file == "mappings.json"

    updated_resources = fmu_dir.sync_dir(new_fmu_dir)

    expected_no_of_resources = 3
    assert len(updated_resources) == expected_no_of_resources
    assert "config" in updated_resources
    assert fmu_dir.config.load().masterdata is None

    assert "_mappings" in updated_resources
    assert len(fmu_dir._mappings.well_mappings) == 0
    assert len(fmu_dir._mappings.stratigraphy_mappings) == 1
    assert fmu_dir._mappings.stratigraphy_mappings[0] == new_strat_mapping

    assert "_changelog" in updated_resources
    updated_changelog: Log[ChangeInfo] = updated_resources["_changelog"]
    expected_log_length = 6

    # Assert existing entries
    assert len(updated_changelog) == expected_log_length
    assert updated_changelog[0].key == "masterdata"
    assert updated_changelog[0].path == fmu_dir.path
    assert updated_changelog[1].key == "stratigraphy"
    assert updated_changelog[1].path == fmu_dir.path

    # Assert merged entry
    assert updated_changelog[2].key == "stratigraphy"
    assert updated_changelog[2].path == new_fmu_dir.path
    assert updated_changelog[2] == new_fmu_dir._changelog.load()[0]

    # Assert entries from merging config and mappings
    assert updated_changelog[3].key == "masterdata"
    assert updated_changelog[3].path == fmu_dir.path
    assert updated_changelog[4].key == "stratigraphy"
    assert updated_changelog[4].path == fmu_dir.path

    # Assert log entry with merge details
    assert updated_changelog[5].change_type == ChangeType.merge
    assert "config" in updated_changelog[5].file
    assert "_changelog" in updated_changelog[5].file
    assert "_mappings" in updated_changelog[5].file


def test_fmu_directory_base_sync_dir_dont_sync_ignored_fields(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that fields to be ignored are not synced.

    When syncing two .fmu directories, Pydantic resources with fields set
    to be ignored in a diff, should not be synced.
    """
    fmu_dir.set_config_value("cache_max_revisions", 5)
    created_at = fmu_dir.config.load().created_at
    created_by = fmu_dir.config.load().created_by
    last_modified_at = fmu_dir.config.load().last_modified_at
    last_modified_by = fmu_dir.config.load().last_modified_by

    new_fmu_dir = extra_fmu_dir
    new_cache_max_revisions = 10
    new_fmu_dir.set_config_value("cache_max_revisions", new_cache_max_revisions)

    updates = fmu_dir.sync_dir(new_fmu_dir)

    assert updates["config"].cache_max_revisions == new_cache_max_revisions

    # created_at and created_by should never change
    assert updates["config"].created_at == created_at
    assert updates["config"].created_by == created_by

    # last_modified_at changes when the config has been updated through the sync
    assert updates["config"].last_modified_at != last_modified_at
    assert updates["config"].last_modified_by == last_modified_by

    expected_log_length = 4
    assert len(updates["_changelog"]) == expected_log_length

    # First entry should be the cache_max_revision update in fmu_dir
    assert updates["_changelog"][0].key == "cache_max_revisions"
    assert "Old value: 5 -> New value: 5" in updates["_changelog"][0].change

    # Second entry should be the cache_max_revision update from the changelog merge
    assert updates["_changelog"][1].key == "cache_max_revisions"
    assert "Old value: 5 -> New value: 10" in updates["_changelog"][1].change
    assert updates["_changelog"][1].path == new_fmu_dir.path

    # Third entry should be the cache_max_revision update from the config merge
    assert updates["_changelog"][2].key == "cache_max_revisions"
    assert "Old value: 5 -> New value: 10" in updates["_changelog"][2].change
    assert updates["_changelog"][2].path == fmu_dir.path

    # Fourth entry should be the logged merge details
    assert updates["_changelog"][3].key == ".fmu"
    assert updates["_changelog"][3].change_type == ChangeType.merge

    # This should not happen, but just to illustrate:
    # Force updating one of the diff ignore fields, adds log entries that will be merged
    new_fmu_dir.set_config_value("created_by", "johndoe")
    updates = fmu_dir.sync_dir(new_fmu_dir)

    # The updated created_by value should not be merged
    assert new_fmu_dir.config.load().created_by == "johndoe"
    assert "config" not in updates
    assert fmu_dir.config.load().created_by != "johndoe"

    # The changelog entry for the update will be merged
    assert updates["_changelog"][4].key == "created_by"
    assert "Old value: user -> New value: johndoe" in updates["_changelog"][4].change


def test_fmu_directory_base_get_dir_diff_with_mappings(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests getting the diff with another .fmu directory with mappings."""
    strat_mapping = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.primary,
        source_id="TopViking",
        target_id="VIKING GP. Top",
    )
    fmu_dir._mappings.update_stratigraphy_mappings(
        StratigraphyMappings(root=[strat_mapping])
    )

    new_fmu_dir = extra_fmu_dir
    new_strat_mapping = copy.deepcopy(strat_mapping)
    new_strat_mapping.source_id = "TopVolantis"
    new_strat_mapping.target_id = "VOLANTIS GP. Top"
    new_fmu_dir._mappings.update_stratigraphy_mappings(
        StratigraphyMappings(root=[new_strat_mapping])
    )

    old_fmu_dir = copy.deepcopy(fmu_dir)
    dir_diff = fmu_dir.get_dir_diff(new_fmu_dir)

    # Assert no changes made to the .fmu directory
    assert fmu_dir.config.load() == old_fmu_dir.config.load()
    assert fmu_dir._changelog.load() == old_fmu_dir._changelog.load()
    assert fmu_dir._mappings.load() == old_fmu_dir._mappings.load()

    # Assert mappings diff is as expected
    assert "_mappings" in dir_diff
    assert len(dir_diff["_mappings"]) == 1
    mappings_diff = dir_diff["_mappings"][0]
    assert mappings_diff[0] == "mappings"
    assert mappings_diff[1] is None
    assert isinstance(mappings_diff[2], Mappings)
    assert len(mappings_diff[2].wells) == 0
    assert len(mappings_diff[2].stratigraphy) == 1
    assert mappings_diff[2].stratigraphy[0] == new_strat_mapping

    # Assert no config changes
    assert "config" in dir_diff
    assert len(dir_diff["config"]) == 0

    # Assert changelog diff has only the mappings update entry
    assert len(dir_diff["_changelog"]) == 1
    changelog_diff = dir_diff["_changelog"][0]
    assert changelog_diff[0] == "changelog"
    assert changelog_diff[1] is None
    assert len(changelog_diff[2]) == 1
    assert changelog_diff[2][0].key == "stratigraphy"
    assert changelog_diff[2][0].file == "mappings.json"


def test_fmu_directory_base_sync_dir_with_mappings(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests syncing mappings resource with another .fmu directory."""
    strat_mapping = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.primary,
        source_id="TopViking",
        target_id="VIKING GP. Top",
    )
    fmu_dir._mappings.update_stratigraphy_mappings(
        StratigraphyMappings(root=[strat_mapping])
    )

    new_fmu_dir = extra_fmu_dir
    new_strat_mapping = copy.deepcopy(strat_mapping)
    new_strat_mapping.source_id = "TopVolantis"
    new_strat_mapping.target_id = "VOLANTIS GP. Top"
    new_fmu_dir._mappings.update_stratigraphy_mappings(
        StratigraphyMappings(root=[new_strat_mapping])
    )

    updates = fmu_dir.sync_dir(new_fmu_dir)

    # Assert mappings are updated as expected
    assert "_mappings" in updates
    assert updates["_mappings"] == fmu_dir._mappings.load()
    assert len(fmu_dir._mappings.well_mappings) == 0
    assert len(fmu_dir._mappings.stratigraphy_mappings) == 1
    assert fmu_dir._mappings.stratigraphy_mappings[0] == new_strat_mapping

    # Assert no config changes
    assert "config" not in updates

    # Assert changelog has all the mapping updates plus the merge entry
    changelog: Log[ChangeInfo] = updates["_changelog"]
    expected_log_length = 4
    assert len(changelog) == expected_log_length
    assert changelog[0].path == fmu_dir.path
    assert changelog[0].key == "stratigraphy"
    assert changelog[1].path == new_fmu_dir.path
    assert changelog[1].key == "stratigraphy"
    assert changelog[2].path == fmu_dir.path
    assert changelog[2].key == "stratigraphy"
    assert changelog[3].path == fmu_dir.path
    assert changelog[3].change_type == ChangeType.merge
    assert "mappings" in changelog[3].file



"""Tests for fmu.resources.config."""

import copy
import json
import uuid
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from fmu.datamodels.common.access import Access
from fmu.datamodels.common.masterdata import (
    Masterdata,
    Smda,
    StratigraphicColumn,
)
from fmu.datamodels.fmu_results.fields import Model

from fmu.settings._fmu_dir import ProjectFMUDirectory, UserFMUDirectory
from fmu.settings._resources.config_managers import (
    ProjectConfigManager,
    UserConfigManager,
)
from fmu.settings.models._enums import ChangeType
from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.models.user_config import UserConfig


@pytest.fixture
def nested_dict() -> dict[str, Any]:
    """A nested dictionary to test get-set dot notation."""
    return {
        "a": 1,
        "b": {
            "c": "2",
            "d": {"e": 3},
        },
    }


def test_config_resource_manager(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests basic facts about the ProjectConfig resource manager."""
    # This manager already exists in 'fmu_dir', but just to
    # try from scratch.
    manager = ProjectConfigManager(fmu_dir)

    assert manager.fmu_dir == fmu_dir
    assert manager.model_class == ProjectConfig
    assert manager._cache is None
    # Resource manager requires this to be implemented
    assert manager.relative_path == Path("config.json")
    assert manager.path == fmu_dir.path / manager.relative_path


def test_user_config_resource_manager(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests basic facts about the ProjectConfig resource manager."""
    # This manager already exists in 'fmu_dir', but just to
    # try from scratch.
    manager = UserConfigManager(user_fmu_dir)

    assert manager.fmu_dir == user_fmu_dir
    assert manager.model_class == UserConfig
    assert manager._cache is None
    # Resource manager requires this to be implemented
    assert manager.relative_path == Path("config.json")
    assert manager.path == user_fmu_dir.path / manager.relative_path


def test_load_config(
    fmu_dir: ProjectFMUDirectory,
    config_model: ProjectConfig,
) -> None:
    """Tests that load() works when a configuration exists."""
    assert fmu_dir.config.load() == config_model
    assert fmu_dir.config._cache == config_model


def test_load_user_config(
    user_fmu_dir: UserFMUDirectory,
    user_config_model: UserConfig,
) -> None:
    """Tests that load() works when a user configuration exists."""
    assert user_fmu_dir.config.load() == user_config_model
    assert user_fmu_dir.config._cache == user_config_model


def test_get_config_missing_file(tmp_path: Path) -> None:
    """Tests get_config raises FileNotFoundError when config.json missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.load()


def test_get_config_invalid_json(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests a corrupted config.json raises ValueError."""
    config_path = fmu_dir.path / "config.json"
    with open(config_path, "a", encoding="utf-8") as f:
        f.write("%")

    with pytest.raises(
        ValueError, match="Invalid JSON in resource file for 'ProjectConfigManager'"
    ):
        fmu_dir.config.load(force=True)


def test_get_dot_notation_key(
    nested_dict: dict[str, Any], fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests the get helper function for dot notation works as expected."""
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.c") == "2"
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.d") == {"e": 3}
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.d.e") == 3  # noqa PLR2004
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.z") is None
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.z", "foo") == "foo"


def test_get_key(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests getting a key on configuration."""
    assert fmu_dir.config.get("created_by") == "user"
    # No nested entries exist in config yet.
    assert fmu_dir.config.get("does.not.exist", "foo") == "foo"
    assert fmu_dir.config.get("notreal", "foo") == "foo"


def test_get_key_config_does_not_exist(tmp_path: Path) -> None:
    """Tests getting a key when the config is missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.get("version")


def test_set_dot_notation_key(
    nested_dict: dict[str, Any], fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests the set helper function for dot notation works as expected."""
    assert nested_dict["b"]["c"] == "2"
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.c", "3")
    assert nested_dict["b"]["c"] == "3"

    assert nested_dict["b"]["d"] == {"e": 3}
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.d", {"e": "s"})
    assert nested_dict["b"]["d"] == {"e": "s"}
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.d.e", 3)
    assert nested_dict["b"]["d"] == {"e": 3}

    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.z") is None
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.z", "foo")
    assert nested_dict["b"]["z"] == "foo"

    # Overwrites existing non-dict values
    fmu_dir.config._set_dot_notation_key(nested_dict, "a.b", {})
    assert nested_dict["a"] != 1
    assert nested_dict["a"]["b"] == {}


def test_set_key(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests setting a key on configuration."""
    fmu_dir.config.set("created_by", "user2")
    assert fmu_dir.config.get("created_by") == "user2"

    config_model = fmu_dir.config.load(force=True)
    config_dict = config_model.model_dump()
    assert config_dict["created_by"] == "user2"

    # No nested entries exist in config yet.
    fmu_dir.config.set("does.not.exist", "foo")
    assert fmu_dir.config.get("does.not.exist") is None
    fmu_dir.config.set("notreal", "foo")
    assert fmu_dir.config.get("notreal") is None

    # Does not write invalid values
    config_model = fmu_dir.config.load(force=True)
    config_dict = config_model.model_dump()
    assert config_dict.get("does", None) is None
    assert config_dict.get("notreal", None) is None

    with pytest.raises(
        ValueError,
        match=(
            "Invalid value set for 'ProjectConfigManager' with key 'version', "
            "value '2.0'"
        ),
    ):
        fmu_dir.config.set("version", 2.0)


def test_set_key_config_does_not_exist(tmp_path: Path) -> None:
    """Tests getting a key when the config is missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.set("version", "200.0.0")


def test_update_config(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests setting a key on configuration."""
    fmu_dir.config.update({"created_by": "user2", "version": "200.0.0", "not.real": 0})
    assert fmu_dir.config.get("created_by") == "user2"
    assert fmu_dir.config.get("version") == "200.0.0"
    assert fmu_dir.config.get("not") is None

    bad_updates = {"created_by": {}, "version": "major"}
    with pytest.raises(
        ValueError,
        match="Invalid value set for 'ProjectConfigManager' with updates "
        f"'{bad_updates}'",
    ):
        fmu_dir.config.update(bad_updates)


def test_update_config_writes_to_changelog(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that config updates are written to changelog."""
    fmu_dir.config.update({"created_by": "user2", "version": "200.0.0", "new.field": 0})
    changelog = fmu_dir._changelog.load()
    expected_log_entries = 3
    assert len(changelog) == expected_log_entries
    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].key == "created_by"
    assert changelog[0].file == "config.json"
    assert changelog[2].change_type == ChangeType.add
    assert changelog[2].key == "new.field"
    assert changelog[2].file == "config.json"


def test_set_smda(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests setting the masterdata.smda value in the config."""
    assert fmu_dir.config.get("masterdata.smda") is None
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())
    assert config_on_disk["masterdata"] is None

    fmu_dir.set_config_value("masterdata", masterdata_dict)

    smda_model = Smda.model_validate(masterdata_dict["smda"])
    # Compare to round-trip'd Pydantic model for uuid-str to UUID() obj conversion
    assert fmu_dir.get_config_value("masterdata.smda") == smda_model.model_dump()
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())

    config_on_disk_model = ProjectConfig.model_validate(config_on_disk)
    assert fmu_dir.config._cache is not None
    assert fmu_dir.config._cache.masterdata is not None
    assert fmu_dir.config._cache.masterdata.smda == smda_model
    assert config_on_disk_model == fmu_dir.config._cache


def test_set_config_writes_to_changelog(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests setting a config value writes update to changelog."""
    fmu_dir.set_config_value("masterdata", masterdata_dict)
    changelog = fmu_dir._changelog.load()
    assert changelog[0].key == "masterdata"
    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].file == "config.json"

    field = fmu_dir.get_config_value("masterdata.smda.field")
    new_field = copy.deepcopy(field)
    new_field[0]["identifier"] = "test"

    fmu_dir.set_config_value("masterdata.smda.field", new_field)
    changelog = fmu_dir._changelog.load()
    expected_log_entries = 2
    assert len(changelog) == expected_log_entries
    assert changelog[1].key == "masterdata.smda.field"
    assert changelog[1].change_type == ChangeType.update
    assert changelog[1].file == "config.json"


def test_set_model_invalid_fails(
    fmu_dir: ProjectFMUDirectory, model_dict: dict[str, Any]
) -> None:
    """Tests setting the model value in the config using an invalid dictionary."""
    assert fmu_dir.config.get("model") is None

    # drop model.name to test validation
    model_dict.pop("name")

    with pytest.raises(ValueError, match="model.name"):
        fmu_dir.set_config_value("model", model_dict)


def test_set_model(fmu_dir: ProjectFMUDirectory, model_dict: dict[str, Any]) -> None:
    """Tests setting the model value in the config."""
    assert fmu_dir.config.get("model") is None
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())
    assert config_on_disk["model"] is None

    fmu_dir.set_config_value("model", model_dict)

    model = Model.model_validate(model_dict)

    assert fmu_dir.get_config_value("model") == model
    assert fmu_dir.get_config_value("model.revision") == "21.0.0"
    assert fmu_dir.get_config_value("model.name") == "Drogon"

    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())

    config_on_disk_model = ProjectConfig.model_validate(config_on_disk)
    assert fmu_dir.config._cache is not None
    assert fmu_dir.config._cache.model == model
    assert config_on_disk_model == fmu_dir.config._cache


def test_set_access_invalid_fails(
    fmu_dir: ProjectFMUDirectory, access_dict: dict[str, Any]
) -> None:
    """Tests setting the access value in the config using an invalid dictionary."""
    assert fmu_dir.config.get("access") is None

    # drop access.asset to test validation
    access_dict.pop("asset")

    with pytest.raises(ValueError, match="access.asset"):
        fmu_dir.set_config_value("access", access_dict)


def test_set_access(fmu_dir: ProjectFMUDirectory, access_dict: dict[str, Any]) -> None:
    """Tests setting the access value in the config."""
    assert fmu_dir.config.get("access") is None
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())
    assert config_on_disk["access"] is None

    fmu_dir.set_config_value("access", access_dict)

    access = Access.model_validate(access_dict)

    assert fmu_dir.get_config_value("access") == access
    assert fmu_dir.get_config_value("access.classification") == "internal"
    assert fmu_dir.get_config_value("access.asset.name") == "Drogon"

    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())

    config_on_disk_model = ProjectConfig.model_validate(config_on_disk)
    assert fmu_dir.config._cache is not None
    assert fmu_dir.config._cache.access == access
    assert config_on_disk_model == fmu_dir.config._cache


def test_update_config_when_it_does_not_exist(tmp_path: Path) -> None:
    """Tests getting a key when the config is missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.update({"created_by": "user", "version": "200.0.0"})


def test_save(fmu_dir: ProjectFMUDirectory, config_dict: dict[str, Any]) -> None:
    """Tests that save functions as expected."""
    config = fmu_dir.config.load()
    assert fmu_dir.config._cache == config
    assert config.created_by == "user"

    config_dict["created_by"] = "user2"
    new_config = ProjectConfig.model_validate(config_dict)
    fmu_dir.config.save(new_config)

    assert fmu_dir.config._cache is not None
    assert fmu_dir.config._cache.created_by == new_config.created_by
    with open(fmu_dir.config.path, encoding="utf-8") as f:
        new_config_dict = json.loads(f.read())

    # Patch the raw datetime string from json
    new_config_dict["created_at"] = datetime.fromisoformat(
        new_config_dict["created_at"]
    )
    new_config_dict["last_modified_at"] = datetime.fromisoformat(
        new_config_dict["last_modified_at"]
    )

    assert new_config_dict["last_modified_at"] is not None
    assert new_config_dict["last_modified_by"] is not None

    config_dict["last_modified_at"] = new_config_dict["last_modified_at"]
    config_dict["last_modified_by"] = new_config_dict["last_modified_by"]
    assert config_dict == new_config_dict


def test_project_config_diff_with_other_config(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    masterdata_dict: dict[str, Any],
    config_model: ProjectConfig,
) -> None:
    """Tests getting the diff between the project config and another config resource."""
    current_cache_max_revisions = 5
    assert fmu_dir.config.load().cache_max_revisions == current_cache_max_revisions

    incoming_config = ProjectConfigManager(extra_fmu_dir)
    new_cache_max_revisions = 10
    config_model.cache_max_revisions = new_cache_max_revisions
    incoming_config.save(config_model)

    diff = fmu_dir.config.get_resource_diff(incoming_config)
    assert len(diff) == 1
    assert diff == [("cache_max_revisions", 5, 10)]

    masterdata: Masterdata = Masterdata.model_validate(masterdata_dict)
    config_with_masterdata = copy.deepcopy(config_model)
    config_with_masterdata.masterdata = masterdata
    incoming_config.save(config_with_masterdata)

    diff = fmu_dir.config.get_resource_diff(incoming_config)
    expected_length = 2
    assert len(diff) == expected_length
    assert diff[0] == ("masterdata", None, masterdata)
    assert diff[1] == ("cache_max_revisions", 5, 10)

    fmu_dir.config.save(config_with_masterdata)

    config_with_strat_col = copy.deepcopy(config_with_masterdata)
    stratigraphic_column = StratigraphicColumn(identifier="test_id", uuid=uuid.uuid4())
    assert config_with_strat_col.masterdata is not None
    config_with_strat_col.masterdata.smda.stratigraphic_column = stratigraphic_column
    incoming_config.save(config_with_strat_col)

    diff = fmu_dir.config.get_resource_diff(incoming_config)
    current_config = fmu_dir.config.load()
    assert current_config.masterdata is not None
    current_strat_col = current_config.masterdata.smda.stratigraphic_column
    expected_length = 2
    assert len(diff) == expected_length
    assert diff[0] == (
        "masterdata.smda.stratigraphic_column.identifier",
        current_strat_col.identifier,
        stratigraphic_column.identifier,
    )
    assert diff[1] == (
        "masterdata.smda.stratigraphic_column.uuid",
        current_strat_col.uuid,
        stratigraphic_column.uuid,
    )


def test_project_config_diff_with_other_config_ignore_fields(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    config_model: ProjectConfig,
) -> None:
    """Tests that the expected fields are ignored when getting the config diff."""
    incoming_config = ProjectConfigManager(extra_fmu_dir)

    # Make sure fields have different values
    incoming_config_model = config_model
    incoming_config_model.created_at = datetime.now(UTC)
    incoming_config_model.created_by = "new_test_user"
    incoming_config_model.last_modified_at = datetime.now(UTC)
    incoming_config_model.last_modified_by = "new_test_user"
    incoming_config.save(incoming_config_model)

    # Assert that fields have different values
    assert incoming_config.load().created_at != fmu_dir.config.load().created_at
    assert incoming_config.load().created_by != fmu_dir.config.load().created_by
    assert (
        incoming_config.load().last_modified_at
        != fmu_dir.config.load().last_modified_at
    )
    assert (
        incoming_config.load().last_modified_by
        != fmu_dir.config.load().last_modified_by
    )

    diff = fmu_dir.config.get_resource_diff(incoming_config)

    # Assert that ignored fields are not included in diff
    assert len(diff) == 0


def test_project_config_merge_with_other_config(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    masterdata_dict: dict[str, Any],
    config_model: ProjectConfig,
) -> None:
    """Tests merging the project config with another config resource."""
    current_cache_max_revisions = 5
    assert fmu_dir.config.load().cache_max_revisions == current_cache_max_revisions

    incoming_config = ProjectConfigManager(extra_fmu_dir)
    new_cache_max_revisions = 10
    config_model.cache_max_revisions = new_cache_max_revisions
    incoming_config.save(config_model)

    updated_resource = fmu_dir.config.merge_resource(incoming_config)
    assert fmu_dir.config.load() == updated_resource
    assert updated_resource.cache_max_revisions == new_cache_max_revisions

    masterdata: Masterdata = Masterdata.model_validate(masterdata_dict)
    config_with_masterdata = copy.deepcopy(config_model)
    config_with_masterdata.masterdata = masterdata
    assert fmu_dir.config.load().masterdata is None
    incoming_config.save(config_with_masterdata)

    updated_resource = fmu_dir.config.merge_resource(incoming_config)
    assert fmu_dir.config.load().masterdata
    assert fmu_dir.config.load() == updated_resource
    assert updated_resource.masterdata == masterdata

    config_with_strat_col = copy.deepcopy(config_with_masterdata)
    stratigraphic_column = StratigraphicColumn(identifier="test_id", uuid=uuid.uuid4())
    assert config_with_strat_col.masterdata is not None
    config_with_strat_col.masterdata.smda.stratigraphic_column = stratigraphic_column
    incoming_config.save(config_with_strat_col)

    updated_resource = fmu_dir.config.merge_resource(incoming_config)
    assert fmu_dir.config.load() == updated_resource
    assert updated_resource.masterdata is not None
    assert updated_resource.masterdata.smda.stratigraphic_column == stratigraphic_column


def test_project_config_reset_has_none_last_modified_fields() -> None:
    """Tests that reset() creates a config with None for last_modified fields."""
    with patch(
        "fmu.settings.models.project_config.getpass.getuser",
        return_value="test_user",
    ):
        config = ProjectConfig.reset()

    assert config.last_modified_at is None
    assert config.last_modified_by is None
    assert config.created_by == "test_user"
    assert config.created_at is not None


def test_project_config_last_modified_set_and_updated_on_save(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that last_modified fields are set and updated on each save."""
    fmu_dir.config.set("cache_max_revisions", 10)
    first_config = fmu_dir.config.load()

    assert first_config.last_modified_at is not None
    assert first_config.last_modified_by is not None

    now = datetime.now(UTC)
    one_min_ago = now - timedelta(minutes=1)
    assert one_min_ago <= first_config.last_modified_at <= now

    fmu_dir.config.set("cache_max_revisions", 15)
    second_config = fmu_dir.config.load()

    assert second_config.last_modified_at is not None
    assert second_config.last_modified_at >= first_config.last_modified_at



"""Tests for fmu.settings.resources.managers."""

import json
import shutil
from datetime import UTC, datetime
from pathlib import Path
from typing import Self
from unittest.mock import patch

import pytest
from pydantic import AwareDatetime, BaseModel

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.lock_manager import LockManager
from fmu.settings._resources.pydantic_resource_manager import (
    MutablePydanticResourceManager,
    PydanticResourceManager,
)
from fmu.settings.types import ResettableBaseModel


class PydanticResourceTest(BaseModel):
    """A test class for a Pydantic resource."""

    foo: str


class MutablePydanticResourceTest(ResettableBaseModel):
    """A test class for a mutable pydantic resource."""

    foo: str


class PydanticManagerTest(PydanticResourceManager[PydanticResourceTest]):
    """A test Pydantic resource manager."""

    def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
        """Initializer."""
        super().__init__(fmu_dir, PydanticResourceTest)

    @property
    def relative_path(self: Self) -> Path:
        """Relative path."""
        return Path("foo.json")


class MutablePydanticManagerTest(
    MutablePydanticResourceManager[MutablePydanticResourceTest]
):
    """A test mutable Pydantic resource manager."""

    def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
        """Initializer."""
        super().__init__(fmu_dir, MutablePydanticResourceTest)

    @property
    def relative_path(self: Self) -> Path:
        """Relative path."""
        return Path("mutable_foo.json")


def test_pydantic_resource_manager_implementation(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that derived classes must implement 'relative_path'."""

    class Manager(PydanticResourceManager[PydanticResourceTest]):
        """A test Pydantic resource manager."""

        def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
            """Initializer."""
            super().__init__(fmu_dir, PydanticResourceTest)

    manager = Manager(fmu_dir)
    with pytest.raises(NotImplementedError):
        _ = manager.relative_path


def test_pydantic_resource_manager_init(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that initialization of a Pydantic resource manager is as expected."""
    test_manager = PydanticManagerTest(fmu_dir)
    assert test_manager.fmu_dir == fmu_dir
    assert test_manager.model_class == PydanticResourceTest

    resource_path = fmu_dir.path / "foo.json"
    assert test_manager.path == resource_path
    assert test_manager.exists is False
    assert test_manager._cache is None

    with pytest.raises(
        FileNotFoundError,
        match=(
            f"Resource file for 'PydanticManagerTest' not found at: '{resource_path}'"
        ),
    ):
        test_manager.load()


def test_pydantic_resource_manager_save(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests saving a Pydantic resource that does not yet exist."""
    test_manager = PydanticManagerTest(fmu_dir)
    resource_model = PydanticResourceTest(foo="bar")

    test_manager.save(resource_model)

    assert test_manager.exists
    assert test_manager._cache == resource_model
    with open(test_manager.path, encoding="utf-8") as f:
        a_dict = json.loads(f.read())

    assert resource_model == PydanticResourceTest.model_validate(a_dict)


def test_pydantic_resource_manager_save_raises_when_locked(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests saving raises when another process holds the lock."""
    lock = LockManager(fmu_dir)
    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.acquire()

    test_manager = PydanticManagerTest(fmu_dir)
    with pytest.raises(PermissionError, match="Cannot write to .fmu directory"):
        test_manager.save(PydanticResourceTest(foo="bar"))

    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.release()


def test_pydantic_resource_manager_load(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests loading a Pydantic resource."""
    test_manager = PydanticManagerTest(fmu_dir)
    test_resource = PydanticResourceTest(foo="bar")
    test_manager.save(test_resource)
    assert test_manager.load() == test_resource
    assert test_manager._cache == test_resource


def test_pydantic_resource_manager_load_force_true(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource with force=True."""
    test_manager = PydanticManagerTest(fmu_dir)
    test_resource = PydanticResourceTest(foo="bar")
    test_manager.save(test_resource)
    assert test_manager.load() == test_resource
    assert test_manager._cache == test_resource

    shadow_test_manager = PydanticManagerTest(fmu_dir)
    shadow_test_resource = PydanticResourceTest(foo="baz")
    shadow_test_manager.save(shadow_test_resource)

    assert shadow_test_manager.load() == shadow_test_resource
    assert shadow_test_manager._cache == shadow_test_resource

    assert test_manager.load() == test_resource
    assert test_manager._cache == test_resource

    assert test_manager.load(force=True) == shadow_test_resource
    assert test_manager._cache == shadow_test_resource


def test_pydantic_resource_manager_load_cache_false(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource with cache=False."""
    test_manager = PydanticManagerTest(fmu_dir)
    test_resource = PydanticResourceTest(foo="bar")
    test_manager.save(test_resource)

    shadow_test_manager = PydanticManagerTest(fmu_dir)
    assert shadow_test_manager.load(store_cache=False) == test_resource
    assert shadow_test_manager._cache is None

    assert shadow_test_manager.load(store_cache=True) == test_resource
    assert shadow_test_manager._cache == test_resource


def test_pydantic_resource_manager_load_force_true_cache_false(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource with force=True, cache=False."""
    test_manager = PydanticManagerTest(fmu_dir)
    test_resource = PydanticResourceTest(foo="bar")
    test_manager.save(test_resource)

    shadow_test_manager = PydanticManagerTest(fmu_dir)
    shadow_test_resource = PydanticResourceTest(foo="baz")
    shadow_test_manager.save(shadow_test_resource)

    assert test_manager.load(force=True, store_cache=False) == shadow_test_resource
    assert test_manager._cache is test_resource


def test_pydantic_resource_manager_loads_invalid_model(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource."""
    test_manager = PydanticManagerTest(fmu_dir)
    test_resource = PydanticResourceTest(foo="bar")
    test_manager.save(test_resource)

    test_resource_dict = test_resource.model_dump()
    test_resource_dict["foo"] = 0

    fmu_dir.write_text_file(test_manager.path, json.dumps(test_resource_dict))

    with pytest.raises(
        ValueError, match=r"Invalid content in resource file[\s\S]*input_value=0"
    ):
        test_manager.load(force=True)


def test_pydantic_resource_manager_save_does_not_cache_when_disabled(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Saving without cache enabled should not create cache artifacts."""
    test_manager = PydanticManagerTest(fmu_dir)
    original_default = test_manager.automatic_caching
    test_manager.automatic_caching = False
    cache_root = fmu_dir.path / "cache"
    try:
        if cache_root.exists():
            shutil.rmtree(cache_root)
        test_manager.save(PydanticResourceTest(foo="bar"))
    finally:
        test_manager.automatic_caching = original_default

    assert not cache_root.exists()


def test_pydantic_resource_manager_save_stores_revision_when_enabled(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Saving with cache enabled should persist a revision snapshot."""
    test_manager = PydanticManagerTest(fmu_dir)
    model = PydanticResourceTest(foo="bar")
    test_manager.save(model)

    cache_root = fmu_dir.path / "cache"
    assert cache_root.is_dir()
    tag_path = cache_root / "CACHEDIR.TAG"
    assert tag_path.read_text(encoding="utf-8").startswith(
        "Signature: 8a477f597d28d172789f06886806bc55"
    )

    config_cache = cache_root / "foo"
    snapshots = list(config_cache.iterdir())
    assert len(snapshots) == 1
    snapshot = snapshots[0]
    assert snapshot.suffix == ".json"
    assert json.loads(snapshot.read_text(encoding="utf-8")) == model.model_dump()


def test_pydantic_resource_manager_revision_cache_trims_excess(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Revision caching should retain only the configured number of snapshots."""
    original_limit = fmu_dir.cache_max_revisions
    fmu_dir.cache_max_revisions = 5
    try:
        test_manager = PydanticManagerTest(fmu_dir)
        test_manager.save(PydanticResourceTest(foo="one"))
        test_manager.save(PydanticResourceTest(foo="two"))
        test_manager.save(PydanticResourceTest(foo="three"))
        test_manager.save(PydanticResourceTest(foo="four"))
        test_manager.save(PydanticResourceTest(foo="five"))
        test_manager.save(PydanticResourceTest(foo="six"))
    finally:
        fmu_dir.cache_max_revisions = original_limit

    config_cache = fmu_dir.path / "cache" / "foo"
    snapshots = sorted(p.name for p in config_cache.iterdir())
    assert len(snapshots) == 5  # noqa: PLR2004

    contents = [
        json.loads((config_cache / name).read_text(encoding="utf-8"))["foo"]
        for name in snapshots
    ]
    assert contents == ["two", "three", "four", "five", "six"]


def test_pydantic_resource_manager_get_resource_diff(
    fmu_dir: ProjectFMUDirectory, extra_fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests the happy path of getting a diff with another Pydantic resource."""
    current_resource = PydanticManagerTest(fmu_dir)
    current_resource.save(PydanticResourceTest(foo="current_value"))
    incoming_resource = PydanticManagerTest(extra_fmu_dir)
    incoming_resource.save(PydanticResourceTest(foo="incoming_value"))

    diff = current_resource.get_resource_diff(incoming_resource)

    assert len(diff) == 1
    assert diff[0][0] == "foo"
    assert diff[0][1] == "current_value"
    assert diff[0][2] == "incoming_value"


def test_pydantic_resource_manager_get_resource_diff_raises_when_other_type(
    fmu_dir: ProjectFMUDirectory, extra_fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests that a TypeError is raised when diffing a different resource type."""
    current_resource = PydanticManagerTest(fmu_dir)
    current_resource.save(PydanticResourceTest(foo="current_value"))

    incoming_resource = MutablePydanticManagerTest(extra_fmu_dir)
    incoming_resource.save(MutablePydanticResourceTest(foo="incoming_value"))

    with pytest.raises(
        TypeError,
        match=(
            "Resources to diff must be of the same type. Current resource is of type "
            "'PydanticResourceTest', incoming resource of type "
            "'MutablePydanticResourceTest'."
        ),
    ):
        current_resource.get_resource_diff(incoming_resource)  # type: ignore


def test_pydantic_resource_manager_get_resource_diff_raises_when_no_file(
    fmu_dir: ProjectFMUDirectory, extra_fmu_dir: ProjectFMUDirectory
) -> None:
    """FileNotFoundError is raised when one of the resources to diff doesn't exits.

    When trying to diff two resources, the resource must
    exist in both directories in order to make a diff.
    """
    current_resource = PydanticManagerTest(fmu_dir)
    incoming_resource = PydanticManagerTest(extra_fmu_dir)
    expected_exc_msg = (
        "Resources to diff must exist in both directories: "
        "Current resource foo.json exists: {}. "
        "Incoming resource foo.json exists: {}."
    )

    with pytest.raises(
        FileNotFoundError, match=expected_exc_msg.format("False", "False")
    ):
        current_resource.get_resource_diff(incoming_resource)

    incoming_resource.save(PydanticResourceTest(foo="incoming_value"))
    with pytest.raises(
        FileNotFoundError, match=expected_exc_msg.format("False", "True")
    ):
        current_resource.get_resource_diff(incoming_resource)

    with pytest.raises(
        FileNotFoundError, match=expected_exc_msg.format("True", "False")
    ):
        incoming_resource.get_resource_diff(current_resource)


def test_pydantic_resource_manager_get_model_diff(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests the happy path of getting a diff between two Pydantic models."""
    test_manager = PydanticManagerTest(fmu_dir)
    current_model = PydanticResourceTest(foo="current_value")
    incoming_model = PydanticResourceTest(foo="incoming_value")

    model_diff = test_manager.get_model_diff(current_model, incoming_model)

    assert len(model_diff) == 1
    assert model_diff[0][0] == "foo"
    assert model_diff[0][1] == "current_value"
    assert model_diff[0][2] == "incoming_value"


def test_pydantic_resource_manager_get_model_diff_raises_when_different_type(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """A ValueError should be raised for a diff between two different model types."""
    test_manager = PydanticManagerTest(fmu_dir)
    current_model = PydanticResourceTest(foo="current_value")
    incoming_model = MutablePydanticResourceTest(foo="incoming_value")

    with pytest.raises(
        ValueError,
        match=(
            "Models must be of the same type. Current model is of type "
            "'PydanticResourceTest', incoming model of type "
            "'MutablePydanticResourceTest'."
        ),
    ):
        test_manager.get_model_diff(current_model, incoming_model)


def test_pydantic_resource_manager_get_model_diff_ignore_fields(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that fields are ignored in model diff when property is set."""

    class ExamplePydanticModel(BaseModel):
        data: str
        metadata: str
        created_at: AwareDatetime = datetime.now(UTC)
        created_by: str = "tester"
        last_modified_at: AwareDatetime | None = None
        last_modified_by: str | None = None

    # Scenario 1: Property not set => no fields ignored
    test_manager = PydanticManagerTest(fmu_dir)
    current_model = ExamplePydanticModel(data="some_data", metadata="some_metadata")
    incoming_model = ExamplePydanticModel(
        data="some_newer_data",
        metadata="some_newer_metadata",
        last_modified_at=datetime.now(UTC),
        last_modified_by="other_tester",
    )
    model_diff = test_manager.get_model_diff(current_model, incoming_model)

    expected_length = 4
    assert len(model_diff) == expected_length
    assert model_diff[0][0] == "data"
    assert model_diff[1][0] == "metadata"
    assert model_diff[2][0] == "last_modified_at"
    assert model_diff[3][0] == "last_modified_by"

    # Scenario 2: Property set => fields ignored
    class PydanticManagerTestIgnore(PydanticManagerTest):
        @property
        def diff_ignore_fields(self: Self) -> list[str]:
            return ["last_modified_at", "last_modified_by"]

    test_manager = PydanticManagerTestIgnore(fmu_dir)
    model_diff = test_manager.get_model_diff(current_model, incoming_model)

    expected_length = 2
    assert len(model_diff) == expected_length
    assert model_diff[0][0] == "data"
    assert model_diff[1][0] == "metadata"


def test_pydantic_resource_manager_get_diff_when_base_model_diff(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests getting a diff between two Pydantic models containing base models.

    Checks that a diff in a base model within the model is returned as expected.
    """

    class ExamplePydanticModel(BaseModel):
        data: str
        metadata: str
        name: str

    class ExampleNestedPydanticModel(BaseModel):
        parent: str
        child: ExamplePydanticModel

    test_manager = PydanticManagerTest(fmu_dir)
    current_model = ExampleNestedPydanticModel(
        parent="parent",
        child=ExamplePydanticModel(
            data="test_data", metadata="test_metadata", name="first"
        ),
    )
    incoming_model = ExampleNestedPydanticModel(
        parent="parent",
        child=ExamplePydanticModel(
            data="test_data", metadata="test_metadata_update", name="second"
        ),
    )
    model_diff = test_manager.get_model_diff(current_model, incoming_model)
    expected_length = 2
    assert len(model_diff) == expected_length
    assert model_diff[0] == ("child.metadata", "test_metadata", "test_metadata_update")
    assert model_diff[1] == ("child.name", "first", "second")


def test_pydantic_resource_manager_get_diff_when_list_diff(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests getting a diff between two Pydantic models containing list values.

    Checks that a diff in a list is returned as expected.
    """

    class ExamplePydanticModel(BaseModel):
        data: str
        name_list: list[str]

    test_manager = PydanticManagerTest(fmu_dir)
    current_list = ["item1", "item2", "item3"]
    current_model = ExamplePydanticModel(data="some_data", name_list=current_list)
    incoming_list = ["item4", "item5", "item3"]
    incoming_model = ExamplePydanticModel(data="some_data", name_list=incoming_list)

    model_diff = test_manager.get_model_diff(current_model, incoming_model)
    assert len(model_diff) == 1
    assert model_diff[0] == ("name_list", current_list, incoming_list)


def test_pydantic_resource_manager_get_diff_when_value_is_none(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests getting a diff between two Pydantic models containing None values.

    Checks that None values in any of the models are handled as expected.
    """

    class ExamplePydanticModel(BaseModel):
        data: str | None

    test_manager = PydanticManagerTest(fmu_dir)
    current_model = ExamplePydanticModel(data=None)
    incoming_model = ExamplePydanticModel(data="new_data")

    model_diff = test_manager.get_model_diff(current_model, incoming_model)
    assert model_diff == [("data", None, "new_data")]

    model_diff = test_manager.get_model_diff(incoming_model, current_model)
    assert model_diff == [("data", "new_data", None)]


def test_mutable_resource_manager_merge_other_resource(
    fmu_dir: ProjectFMUDirectory, extra_fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests merging a Pydantic resource into the current resource."""
    current_resource = MutablePydanticManagerTest(fmu_dir)
    current_resource.save(MutablePydanticResourceTest(foo="current_value"))
    incoming_resource = MutablePydanticManagerTest(extra_fmu_dir)
    incoming_resource.save(MutablePydanticResourceTest(foo="incoming_value"))

    updated_resource = current_resource.merge_resource(incoming_resource)
    assert updated_resource == incoming_resource.load()


def test_mutable_resource_manager_merge_resource_raises_when_other_type(
    fmu_dir: ProjectFMUDirectory, extra_fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests that a TypeError is raised when merging a different resource type."""
    current_resource = MutablePydanticManagerTest(fmu_dir)
    current_resource.save(MutablePydanticResourceTest(foo="current_value"))
    incoming_resource = PydanticManagerTest(extra_fmu_dir)
    incoming_resource.save(PydanticResourceTest(foo="incoming_value"))

    with pytest.raises(
        TypeError,
        match=(
            "Merging pydantic resource failed. The incoming resource must be of type "
            "'MutablePydanticResourceTest'. The provided model was of type "
            "'PydanticResourceTest'."
        ),
    ):
        current_resource.merge_resource(incoming_resource)  # type: ignore



"""The model for config.json."""

import getpass
from datetime import UTC, datetime
from pathlib import Path
from typing import Literal, Self

from pydantic import AwareDatetime, BaseModel, Field

from fmu.datamodels.common.access import Access
from fmu.datamodels.common.masterdata import Masterdata
from fmu.datamodels.fmu_results.fields import Model
from fmu.settings import __version__
from fmu.settings.types import ResettableBaseModel, VersionStr  # noqa: TC001


class RmsCoordinateSystem(BaseModel):
    """The project coordinate system of an RMS project."""

    name: str
    """Name of the coordinate system."""


class RmsStratigraphicZone(BaseModel):
    """A stratigraphic zone from an RMS project."""

    name: str
    """Name of the zone."""

    top_horizon_name: str
    """Name of the horizon at the top of the zone."""

    base_horizon_name: str
    """Name of the horizon at the base of the zone."""

    stratigraphic_column_name: list[str] | None = None
    """Name of the stratigraphic column the zone belongs to."""


class RmsHorizon(BaseModel):
    """A horizon from an RMS project."""

    name: str
    """Name of the horizon."""

    type: Literal[
        "calculated",
        "calculated_unconformity",
        "interpreted",
        "interpreted_unconformity",
        "interpreted_intrusion",
    ]
    """Type of the horizon."""


class RmsWell(BaseModel):
    """A well from an RMS project."""

    name: str
    """Name of the well."""


class RmsProject(BaseModel):
    """RMS project configuration."""

    path: Path
    version: str
    coordinate_system: RmsCoordinateSystem | None = None
    zones: list[RmsStratigraphicZone] | None = None
    horizons: list[RmsHorizon] | None = None
    wells: list[RmsWell] | None = None


class ProjectConfig(ResettableBaseModel):
    """The configuration file in a .fmu directory.

    Stored as config.json.
    """

    version: VersionStr
    created_at: AwareDatetime
    created_by: str
    last_modified_at: AwareDatetime | None = None
    last_modified_by: str | None = None
    masterdata: Masterdata | None = None
    model: Model | None = None
    access: Access | None = None
    cache_max_revisions: int = Field(default=5, ge=5)
    rms: RmsProject | None = None

    @classmethod
    def reset(cls: type[Self]) -> Self:
        """Resets the configuration to defaults.

        Returns:
            The new default Config object
        """
        return cls(
            version=__version__,
            created_at=datetime.now(UTC),
            created_by=getpass.getuser(),
            last_modified_at=None,
            last_modified_by=None,
            masterdata=None,
            model=None,
            access=None,
            cache_max_revisions=5,
            rms=None,
        )



"""Model for the mappings.json file."""

from typing import Any

from pydantic import BaseModel, Field, model_validator

from fmu.datamodels.context.mappings import (
    AnyIdentifierMapping,
    DataSystem,
    RelationType,
    StratigraphyMappings,
)


class Mappings(BaseModel):
    """Represents the mappings file in a .fmu directory."""

    stratigraphy: StratigraphyMappings = Field(
        default_factory=lambda: StratigraphyMappings(root=[])
    )
    """Stratigraphy mappings in the mappings file."""

    # Todo: Add wells model
    wells: list[Any] = Field(default_factory=list)
    """Well mappings in the mappings file."""


class MappingGroup(BaseModel):
    """A mapping group containing a primary mapping and its related mappings.

    This is a view of the data for GUI display purposes, not how it's stored.
    Groups all mappings (primary, aliases, equivalents) that share the same target.

    Validates that all mappings share the group target_id, target_system,
    and source_system.
    """

    target_id: str
    target_system: DataSystem
    source_system: DataSystem
    mappings: list[AnyIdentifierMapping]

    @property
    def primary(self) -> AnyIdentifierMapping | None:
        """Get the primary mapping.

        Takes the first primary, which should be the only one.
        """
        primaries = [
            m for m in self.mappings if m.relation_type == RelationType.primary
        ]
        return primaries[0] if primaries else None

    @property
    def aliases(self) -> list[AnyIdentifierMapping]:
        """Get all alias mappings."""
        return [m for m in self.mappings if m.relation_type == RelationType.alias]

    @property
    def equivalents(self) -> list[AnyIdentifierMapping]:
        """Get all equivalent mappings."""
        return [m for m in self.mappings if m.relation_type == RelationType.equivalent]

    @model_validator(mode="after")
    def validate_group(self) -> "MappingGroup":
        """Ensure mappings align with the group target/source and primary count."""
        if not self.mappings:
            return self

        primary_count = sum(
            1
            for mapping in self.mappings
            if mapping.relation_type == RelationType.primary
        )
        if primary_count > 1:
            raise ValueError("MappingGroup must contain at most one primary mapping.")

        for mapping in self.mappings:
            if mapping.target_id != self.target_id:
                raise ValueError(
                    "All mappings in MappingGroup must share target_id "
                    f"'{self.target_id}'."
                )
            if mapping.target_system != self.target_system:
                raise ValueError(
                    "All mappings in MappingGroup must share target_system "
                    f"'{self.target_system}'."
                )
            if mapping.source_system != self.source_system:
                raise ValueError(
                    "All mappings in MappingGroup must share source_system "
                    f"'{self.source_system}'."
                )
        return self

    def to_display_dict(self) -> dict[str, Any]:
        """Convert to a dictionary for display."""
        return {
            "official_name": self.target_id,
            "target_system": self.target_system.value,
            "source_system": self.source_system.value,
            "primary_source": self.primary.source_id if self.primary else None,
            "aliases": [m.source_id for m in self.aliases],
            "equivalents": [m.source_id for m in self.equivalents],
        }



from __future__ import annotations

import copy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Self

from fmu.settings._resources.pydantic_resource_manager import PydanticResourceManager
from fmu.settings.models.mappings import Mappings

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from fmu.datamodels.context.mappings import (
        StratigraphyMappings,
    )
    from fmu.settings._fmu_dir import (
        FMUDirectoryBase,
    )


class MappingsManager(PydanticResourceManager[Mappings]):
    """Manages the .fmu mappings file."""

    def __init__(self: Self, fmu_dir: FMUDirectoryBase) -> None:
        """Initializes the mappings resource manager."""
        super().__init__(fmu_dir, Mappings)

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the mappings file."""
        return Path("mappings.json")

    @property
    def stratigraphy_mappings(self: Self) -> StratigraphyMappings:
        """Get all stratigraphy mappings."""
        return self.load().stratigraphy

    @property
    def well_mappings(self: Self) -> list[Any]:
        """Get all well mappings."""
        return self.load().wells

    def update_stratigraphy_mappings(
        self: Self, strat_mappings: StratigraphyMappings
    ) -> StratigraphyMappings:
        """Updates the stratigraphy mappings in the mappings resource."""
        mappings: Mappings = self.load() if self.exists else Mappings()

        old_mappings_dict = copy.deepcopy(mappings.model_dump())
        mappings.stratigraphy = strat_mappings
        self.save(mappings)

        self.fmu_dir._changelog.log_update_to_changelog(
            updates={"stratigraphy": mappings.stratigraphy},
            old_resource_dict=old_mappings_dict,
            relative_path=self.relative_path,
        )

        return self.stratigraphy_mappings

    def update_well_mappings(self: Self) -> None:
        # TODO: Add well mappings functionality
        raise NotImplementedError

    def get_mappings_diff(self: Self, incoming_mappings: MappingsManager) -> Mappings:
        """Get mappings diff with the incoming mappings resource.

        All mappings from the incommng mappings resource are returned.
        """
        if self.exists and incoming_mappings.exists:
            return incoming_mappings.load()
        raise FileNotFoundError(
            "Mappings resources to diff must exist in both directories: "
            f"Current mappings resource exists: {self.exists}. "
            f"Incoming mappings resource exists: {incoming_mappings.exists}."
        )

    def merge_mappings(self: Self, incoming_mappings: MappingsManager) -> Mappings:
        """Merge the mappings from the incoming mappings resource.

        The current mappings will be updated with the mappings
        from the incoming resource.
        """
        mappings_diff = self.get_mappings_diff(incoming_mappings)
        return self.merge_changes(mappings_diff)

    def merge_changes(self: Self, changes: Mappings) -> Mappings:
        """Merge the mappings changes into the current mappings.

        The current mappings will be updated with the mappings
        in the change object.
        """
        if len(changes.stratigraphy) > 0 or len(self.stratigraphy_mappings) > 0:
            self.update_stratigraphy_mappings(changes.stratigraphy)
        if len(changes.wells) > 0 or len(self.well_mappings) > 0:
            self.update_well_mappings()
        return self.load()



"""Contains enumerations used in this package."""

from enum import StrEnum


class MappingType(StrEnum):
    """The discriminator used between mappings.

    Each of these types should have their own mapping class derived of some sort of
    mapping.
    """

    fault = "fault"
    stratigraphy = "stratigraphy"
    well = "well"


class RelationType(StrEnum):
    """The kind of relation this mapping represents."""

    alias = "alias"
    child_to_parent = "child_to_parent"
    equivalent = "equivalent"
    fmu_to_target = "fmu_to_target"
    predecessor_to_successor = "predecessor_to_successor"


class DataEntrySource(StrEnum):
    user = "user"
    automated = "automated"


class TargetSystem(StrEnum):
    smda = "smda"


class ChangeType(StrEnum):
    """The types of change that can be made on a file."""

    update = "update"
    remove = "remove"
    add = "add"
    reset = "reset"
    merge = "merge"


class FilterType(StrEnum):
    """The supported types to filter on."""

    date = "date"
    number = "number"
    text = "text"


class CacheResource(StrEnum):
    """Resources that can be cached and restored."""

    config = "config.json"
    mappings = "mappings.json"



"""Tests for ChangelogManager."""

import copy
import uuid
import warnings
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any

import pandas as pd
import pytest
from fmu.datamodels.common.masterdata import StratigraphicColumn

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.changelog_manager import ChangelogManager
from fmu.settings.models._enums import ChangeType, FilterType
from fmu.settings.models.change_info import ChangeInfo
from fmu.settings.models.log import Filter, Log, LogFileName

DATE_TIME_NOW = datetime.now(UTC)


@pytest.fixture
def change_entry() -> ChangeInfo:
    """Returns a valid ChangeInfo object."""
    return ChangeInfo(
        timestamp=DATE_TIME_NOW,
        change_type=ChangeType.add,
        user="test",
        path=Path("/test_folder"),
        file="config.json",
        change="Added new field to smda masterdata.",
        hostname="hostname",
        key="masterdata",
    )


@pytest.fixture
def change_entry_list() -> list[ChangeInfo]:
    """Returns a list of valid ChangeInfo entries."""
    return [
        ChangeInfo(
            timestamp=DATE_TIME_NOW - timedelta(days=2),
            change_type=ChangeType.add,
            user="user_first_entry",
            path=Path("/path_first_entry"),
            file="config.json",
            change="Added new field.",
            hostname="hostname_first_entry",
            key="masterdata",
        ),
        ChangeInfo(
            timestamp=DATE_TIME_NOW - timedelta(days=1),
            change_type=ChangeType.update,
            user="user_second_entry",
            path=Path("/path_second_entry"),
            file="config.json",
            change="Updated field.",
            hostname="hostname_second_entry",
            key="masterdata",
        ),
        ChangeInfo(
            timestamp=DATE_TIME_NOW,
            change_type=ChangeType.remove,
            user="user_third_entry",
            path=Path("/path_third_entry"),
            file="config.json",
            change="Removed field.",
            hostname="hostname_third_entry",
            key="masterdata",
        ),
        ChangeInfo(
            timestamp=DATE_TIME_NOW,
            change_type=ChangeType.reset,
            user="user_fourth_entry",
            path=Path("/path_fourth_entry"),
            file="config.json",
            change="Reset field.",
            hostname="hostname_fourth_entry",
            key="masterdata",
        ),
    ]


def test_changelog_manager_instantiation(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests basic facts about the ChangelogManager."""
    changelog: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog.fmu_dir == fmu_dir
    assert changelog.relative_path == Path("logs") / LogFileName.changelog
    assert changelog.exists is False
    with pytest.raises(
        FileNotFoundError, match="Resource file for 'ChangelogManager' not found"
    ):
        changelog.load()

    assert changelog.model_class == Log[ChangeInfo]


def test_changelog_manager_add_entry(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests adding a new entry to a non-existing changelog.

    This should create the changelog file and add the new entry at the end.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource.exists is False
    with pytest.raises(
        FileNotFoundError, match="Resource file for 'ChangelogManager' not found"
    ):
        changelog_resource.load()

    changelog_resource.add_log_entry(change_entry)

    assert changelog_resource.exists is True
    changelog = changelog_resource.load()
    assert changelog[0] == change_entry

    changelog_resource.add_log_entry(change_entry)
    changelog_resource.add_log_entry(change_entry)
    expected_log_entries = 3

    updated_changelog: Log[ChangeInfo] = changelog_resource.load()
    assert len(updated_changelog) == expected_log_entries


def test_changelog_manager_add_invalid_entry(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests adding an invalid entry to an existing changelog.

    This should raise a Value error and not add the invalid log entry.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    changelog_resource.add_log_entry(change_entry)
    assert changelog_resource.exists

    change_entry_with_issues = copy.deepcopy(change_entry)
    del change_entry_with_issues.change_type

    with pytest.raises(
        ValueError, match=r"Invalid log entry added to 'Log\[ChangeInfo\]' "
    ):
        changelog_resource.add_log_entry(change_entry_with_issues)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    assert len(changelog) == 1
    assert changelog[0] == change_entry


def test_changelog_manager_add_invalid_entry_no_file_created(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests adding an invalid entry to a non-existing changelog.

    This should raise a Value error and not create the changelog file.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource.exists is False

    change_entry_with_issues = copy.deepcopy(change_entry)
    change_entry_with_issues.change_type = "invalid change_type"  # type: ignore

    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        with pytest.raises(
            ValueError, match=r"Invalid log entry added to 'Log\[ChangeInfo\]' "
        ):
            changelog_resource.add_log_entry(change_entry_with_issues)

    assert changelog_resource.exists is False


def test_changelog_filter_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with == operator.

    The filter should return all changelog entries where the value of the field
    `field_name` equals the filter value.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 1
    assert filtered_log[0] == changelog[0]

    filter = Filter(
        field_name="user",
        filter_value="user_second_entry",
        filter_type=FilterType.text,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 1
    assert filtered_log[0] == changelog[1]

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW),
        filter_type=FilterType.date,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp == DATE_TIME_NOW for entry in filtered_log)

    filter = Filter(
        field_name="change",
        filter_value="Changed field.",
        filter_type=FilterType.text,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_not_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with != operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is not equal to the filter value.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 3
    assert len(filtered_log) == expected_log_entries
    assert all(entry.change_type is not ChangeType.add for entry in filtered_log)

    filter = Filter(
        field_name="user",
        filter_value="user_second_entry",
        filter_type=FilterType.text,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 3
    assert len(filtered_log) == expected_log_entries
    assert all(entry.user != "user_second_entry" for entry in filtered_log)

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW),
        filter_type=FilterType.date,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp != DATE_TIME_NOW for entry in filtered_log)

    filter = Filter(
        field_name="key",
        filter_value="masterdata",
        filter_type=FilterType.text,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_less_or_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with <= operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is less or equal to the filter value. Attempts to filter
    strings with the <= operator should raise an exception.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="<=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator <= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    filter = Filter(
        field_name="user",
        filter_value="user_third_entry",
        filter_type=FilterType.text,
        operator="<=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator <= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    yesterday = DATE_TIME_NOW - timedelta(days=1)
    filter = Filter(
        field_name="timestamp",
        filter_value=str(yesterday),
        filter_type=FilterType.date,
        operator="<=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp <= yesterday for entry in filtered_log)

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW - timedelta(days=3)),
        filter_type=FilterType.date,
        operator="<=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_greater_or_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with >= operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is greater or equal to the filter value. Attempts to filter
    strings with the >= operator should raise an exception.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator=">=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator >= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    filter = Filter(
        field_name="user",
        filter_value="user_third_entry",
        filter_type=FilterType.text,
        operator=">=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator >= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    yesterday = DATE_TIME_NOW - timedelta(days=1)
    filter = Filter(
        field_name="timestamp",
        filter_value=str(yesterday),
        filter_type=FilterType.date,
        operator=">=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 3
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp >= yesterday for entry in filtered_log)

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW + timedelta(days=1)),
        filter_type=FilterType.date,
        operator=">=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_greater_than_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with > operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is greater than the filter value. Attempts to filter
    strings with the > operator should raise an exception.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator=">",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator > applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    filter = Filter(
        field_name="user",
        filter_value="user_third_entry",
        filter_type=FilterType.text,
        operator=">",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator > applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    yesterday = DATE_TIME_NOW - timedelta(days=1)
    filter = Filter(
        field_name="timestamp",
        filter_value=str(yesterday),
        filter_type=FilterType.date,
        operator=">",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp > yesterday for entry in filtered_log)

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW),
        filter_type=FilterType.date,
        operator=">",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_less_than_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with < operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is less than the filter value. Attempts to filter
    strings with the < operator should raise an exception.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="<",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator < applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    filter = Filter(
        field_name="user",
        filter_value="user_third_entry",
        filter_type=FilterType.text,
        operator="<",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator < applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    yesterday = DATE_TIME_NOW - timedelta(days=1)
    filter = Filter(
        field_name="timestamp",
        filter_value=str(yesterday),
        filter_type=FilterType.date,
        operator="<",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 1

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW + timedelta(days=1)),
        filter_type=FilterType.date,
        operator="<",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 4
    assert len(filtered_log) == expected_log_entries


def test_changelog_dataframe_cached_after_filtering(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests that _cached_dataframe is set after filtering has been applied."""
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource._cached_dataframe is None
    changelog_resource.add_log_entry(change_entry)
    assert changelog_resource._cached_dataframe is None

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="==",
    )
    changelog_resource.filter_log(filter)
    assert changelog_resource._cached_dataframe is not None


def test_changelog_dataframe_cache_cleared(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests that _cached_dataframe is cleared when the changelog is updated."""
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource._cached_dataframe is None

    changelog_resource._cached_dataframe = pd.DataFrame(["some_data"])
    assert changelog_resource._cached_dataframe is not None

    changelog_resource.add_log_entry(change_entry)
    assert changelog_resource._cached_dataframe is None


def test_log_update_to_changelog_when_flat_dict(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that updates of flat dictionaries are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)

    first_key = "first_key"
    first_value = "first_test_value"
    old_resource_dict = {first_key: first_value, "some_key": "some_value"}

    updated_value = "updated_value"
    added_key = "added_key"
    added_value = "added_value"
    updates: dict[str, Any] = {
        first_key: updated_value,
        added_key: added_value,
    }

    changelog_resource.log_update_to_changelog(
        updates, old_resource_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 2
    assert len(changelog) == expected_log_entries

    expected_change_string = (
        f"Updated field '{first_key}'. Old value: {first_value}"
        f" -> New value: {updated_value}"
    )
    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].change == expected_change_string

    expected_change_string = f"Added field '{added_key}'. New value: {added_value}"
    assert changelog[1].change_type == ChangeType.add
    assert expected_change_string == changelog[1].change


def test_log_update_to_changelog_when_nested_dict(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests that updates of nested dictionaries are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)

    updated_country = [
        {
            "identifier": "Norge",
            "uuid": "00000000-0000-0000-0000-000000000000",
        }
    ]

    first_key = "smda.country"
    new_key = "new_key"
    new_nested_key = "new.nested.key"
    updates: dict[str, Any] = {
        first_key: updated_country,
        new_key: "new_value",
        new_nested_key: "new_nested_value",
    }

    changelog_resource.log_update_to_changelog(
        updates, masterdata_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 3
    assert len(changelog) == expected_log_entries

    expected_old_value = str(masterdata_dict["smda"]["country"])
    expected_change_string = (
        f"Updated field '{first_key}'. Old value: {expected_old_value}"
        f" -> New value: {str(updated_country)}"
    )

    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].change == expected_change_string

    expected_change_string = f"Added field '{new_key}'. New value: new_value"
    assert changelog[1].change_type == ChangeType.add
    assert expected_change_string == changelog[1].change

    expected_change_string = (
        f"Added field '{new_nested_key}'. New value: new_nested_value"
    )
    assert changelog[2].change_type == ChangeType.add
    assert expected_change_string == changelog[2].change


def test_log_update_to_changelog_when_none_values(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that updates with None values are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    first_key = "first_key"
    first_value = "first_test_value"
    second_key = "some_key"
    old_resource_dict = {first_key: first_value, second_key: None}

    updated_value = "updated_value"
    added_key = "added_key"
    updates: dict[str, Any] = {
        first_key: None,
        second_key: updated_value,
        added_key: None,
    }

    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    changelog_resource.log_update_to_changelog(
        updates, old_resource_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 3
    assert len(changelog) == expected_log_entries

    assert changelog[0].change_type == ChangeType.update
    expected_change_string = (
        f"Updated field '{first_key}'. Old value: {first_value}"
        f" -> New value: {str(None)}"
    )
    assert changelog[0].change == expected_change_string

    assert changelog[1].change_type == ChangeType.update
    expected_change_string = (
        f"Updated field '{second_key}'. Old value: {str(None)}"
        f" -> New value: {updated_value}"
    )
    assert changelog[1].change == expected_change_string

    assert changelog[2].change_type == ChangeType.add
    expected_change_string = f"Added field '{added_key}'. New value: {str(None)}"
    assert changelog[2].change_type == ChangeType.add
    assert expected_change_string == changelog[2].change


def test_log_update_to_changelog_when_base_model_values(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests that updates to BaseModel objects are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    strat_column = StratigraphicColumn(identifier="test_strat", uuid=uuid.uuid4())

    test_update = "smda.stratigraphic_column"
    test_add = "new.field"
    updates: dict[str, Any] = {test_update: strat_column, test_add: strat_column}

    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    changelog_resource.log_update_to_changelog(
        updates, masterdata_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 2
    assert len(changelog) == expected_log_entries

    assert changelog[0].key == test_update
    assert changelog[0].change_type == ChangeType.update
    old_value = masterdata_dict["smda"]["stratigraphic_column"]
    expected_change_string = (
        f"Updated field '{test_update}'. Old value: {str(old_value)}"
        f" -> New value: {str(strat_column.model_dump())}"
    )
    assert expected_change_string == changelog[0].change

    assert changelog[1].key == test_add
    assert changelog[1].change_type == ChangeType.add
    expected_change_string = (
        f"Added field '{test_add}'. New value: {str(strat_column.model_dump())}"
    )
    assert expected_change_string == changelog[1].change


def test_changelog_get_latest_change_timestamp(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests that the timestamp of the latest added log entry is returned."""
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    expected_latest_timestamp = change_entry_list[-1].timestamp
    assert (
        changelog_resource._get_latest_change_timestamp() == expected_latest_timestamp
    )


def test_changelog_get_changelog_diff_with_other_changelog(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    change_entry_list: list[ChangeInfo],
) -> None:
    """Tests that the new entries from the incoming changelog are returned.

    When getting a diff between two changelogs, all change entries in the incoming
    changelog newer than the log entries in the current changelog, should be returned.
    """
    current_changelog: ChangelogManager = ChangelogManager(fmu_dir)
    current_changelog.add_log_entry(
        ChangeInfo(
            timestamp=DATE_TIME_NOW - timedelta(days=1),
            change_type=ChangeType.add,
            user="old_test",
            path=Path("/test_folder"),
            file="config.json",
            change="Added new field to smda masterdata one day ago",
            hostname="hostname",
            key="masterdata",
        ),
    )
    incoming_changelog: ChangelogManager = ChangelogManager(extra_fmu_dir)
    for entry in change_entry_list:
        incoming_changelog.add_log_entry(entry)

    diff = current_changelog.get_changelog_diff(incoming_changelog)

    expected_diff_entries = 2
    assert len(diff) == expected_diff_entries
    assert diff[0] == change_entry_list[2]
    assert diff[1] == change_entry_list[3]

    starting_point = current_changelog._get_latest_change_timestamp()
    for entry in diff:
        assert entry.timestamp > starting_point
    assert change_entry_list[0].timestamp <= starting_point
    assert change_entry_list[1].timestamp <= starting_point


def test_changelog_get_changelog_diff_with_old_changelog(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    change_entry: ChangeInfo,
) -> None:
    """Tests that the diff is empty when all incoming changes are old.

    When the latest change in the current changelog is newer than the entries
    in the incoming changelog, the result should be an empty Log object.
    """
    current_changelog: ChangelogManager = ChangelogManager(fmu_dir)
    current_changelog.add_log_entry(change_entry)

    incoming_changelog: ChangelogManager = ChangelogManager(extra_fmu_dir)
    incoming_changelog.add_log_entry(
        ChangeInfo(
            timestamp=DATE_TIME_NOW - timedelta(seconds=2),
            change_type=ChangeType.add,
            user="old_test",
            path=Path("/test_folder"),
            file="config.json",
            change="Added new field to smda masterdata two seconds ago",
            hostname="hostname",
            key="masterdata",
        ),
    )
    incoming_changelog.add_log_entry(
        ChangeInfo(
            timestamp=DATE_TIME_NOW - timedelta(seconds=3),
            change_type=ChangeType.add,
            user="older_test",
            path=Path("/test_folder"),
            file="config.json",
            change="Added new field to smda masterdata three seconds ago.",
            hostname="hostname",
            key="masterdata",
        )
    )

    diff = current_changelog.get_changelog_diff(incoming_changelog)
    assert len(diff) == 0


def test_changelog_get_changelog_diff_with_other_changelog_raises(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    change_entry: ChangeInfo,
) -> None:
    """Exception is raised when any of the changelog resources to diff does not exist.

    When trying to diff two changelog resources, the changelog file must
    exist in both directories in order to make a diff.
    """
    current_changelog: ChangelogManager = ChangelogManager(fmu_dir)
    incoming_changelog: ChangelogManager = ChangelogManager(extra_fmu_dir)

    expected_exc_msg = (
        "Changelog resources to diff must exist in both directories: "
        "Current changelog resource exists: {}. "
        "Incoming changelog resource exists: {}."
    )
    with pytest.raises(
        FileNotFoundError, match=expected_exc_msg.format("False", "False")
    ):
        current_changelog.get_changelog_diff(incoming_changelog)

    current_changelog.add_log_entry(change_entry)
    with pytest.raises(
        FileNotFoundError, match=expected_exc_msg.format("True", "False")
    ):
        current_changelog.get_changelog_diff(incoming_changelog)

    with pytest.raises(
        FileNotFoundError, match=expected_exc_msg.format("False", "True")
    ):
        incoming_changelog.get_changelog_diff(current_changelog)


def test_changelog_merge_changelog_with_other_changelog(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    change_entry_list: list[ChangeInfo],
) -> None:
    """Tests merging a changelog resource with an incoming changelog.

    When merging two changelogs, all log entries from the incoming changelog newer
    than the log entries in the current changelog are added.
    """
    current_changelog: ChangelogManager = ChangelogManager(fmu_dir)
    existing_entry = ChangeInfo(
        timestamp=DATE_TIME_NOW - timedelta(days=1),
        change_type=ChangeType.add,
        user="old_test",
        path=Path("/test_folder"),
        file="config.json",
        change="Added new field to smda masterdata one day ago",
        hostname="hostname",
        key="masterdata",
    )
    current_changelog.add_log_entry(existing_entry)

    incoming_changelog: ChangelogManager = ChangelogManager(extra_fmu_dir)
    for entry in change_entry_list:
        incoming_changelog.add_log_entry(entry)

    updated_changelog = current_changelog.merge_changelog(incoming_changelog)

    expected_entries = 3
    assert len(updated_changelog) == expected_entries
    assert updated_changelog[0] == existing_entry
    assert updated_changelog[1] == change_entry_list[2]
    assert updated_changelog[2] == change_entry_list[3]

    current_log = current_changelog.load(force=True)
    assert updated_changelog == current_log


def test_changelog_merge_changes_into_current_changelog(
    fmu_dir: ProjectFMUDirectory,
    extra_fmu_dir: ProjectFMUDirectory,
    change_entry: ChangeInfo,
    change_entry_list: list[ChangeInfo],
) -> None:
    """Tests that all log entries in the change object are added to the changelog."""
    current_changelog: ChangelogManager = ChangelogManager(fmu_dir)
    current_changelog.add_log_entry(change_entry)

    new_changelog: ChangelogManager = ChangelogManager(extra_fmu_dir)
    for entry in change_entry_list:
        new_changelog.add_log_entry(entry)
    change = change_entry_list
    updated_changelog = current_changelog.merge_changes(change)

    expected_entries = 5
    assert len(updated_changelog) == expected_entries
    assert updated_changelog == current_changelog.load()


def test_changelog_log_merge_to_changelog_merge_details(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that the merge details logged to the changelog is as expected."""
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    source_path = changelog_resource.path.parent.parent
    incoming_path = Path("/some_path/.fmu")
    resources = ["config", "_changelog"]

    changelog_resource.log_merge_to_changelog(
        source_path=source_path, incoming_path=incoming_path, merged_resources=resources
    )

    changelog = changelog_resource.load()
    assert len(changelog) == 1
    merge_entry = changelog[0]
    assert merge_entry.change_type == ChangeType.merge
    assert merge_entry.path == source_path
    assert merge_entry.key == ".fmu"

    expected_change_string = (
        f"Merged resources 'config', '_changelog' from "
        f"'{incoming_path}' into '{source_path}'."
    )
    assert merge_entry.change == expected_change_string
